import numpy as np
from numba import njit
import time

from scipy.ndimage import uniform_filter1d
from tqdm.autonotebook import tqdm
from piscat.InputOutput.cpu_configurations import CPUConfigurations
from piscat.Preproccessing import FPNc
from joblib import Parallel, delayed

from PySide6 import QtGui
from PySide6 import QtWidgets
from PySide6 import QtCore

from PySide6.QtCore import *


@njit(parallel=True)
def numba_divide(ndarray_1, ndarray_2):
    divide_result = np.divide(ndarray_1, ndarray_2)
    return divide_result

@njit(parallel=True)
def numba_sum(video, axis):
    sum_result = np.sum(video, axis=axis)
    return sum_result

@njit(parallel=True)
def numba_diff(ndarray_1, ndarray_2):
    diff_result = ndarray_1 -ndarray_2
    return diff_result


class WorkerSignals(QObject):

    updateProgress_DRA = Signal(int)
    result_final = Signal(object)
    finished_DRA = Signal()
    DRA_complete_signal = Signal(bool)
    updateProgress_FPNC = Signal(int)


class DifferentialRollingAverage(QRunnable):

    def __init__(self, video=None, batchSize=500, flag_GUI=False, object_update_progressBar=None,
                 mode_FPN='mFPN', FPN_flag_GUI=False, gui_select_correction_axis=1):
        """
        Differential Rolling Average (DRA).

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        batchSize: int
            The number of frames in each batch.

        mode_FPN: {‘cpFPN’, ‘mFPN’, ‘wFPN’, 'fFPN'}, optional
            Flag that defines method of FPNc.

                    * `mFPN`: Median fixed pattern noise correction
                    * `cpFPN`: Median fixed pattern noise correction
                    * `wFPN`: Wavelet FPNc
                    * `fFPN`: FFT2D_Wavelet FPNc

        optional_1: GUI
            These flags are used when GUI calls this method.

            * `flag_GUI`: bool
                This flag is defined as True when GUI calls this method.

            * `FPN_flag_GUI`: bool
                This flag is defined as True when GUI calls this method while we want activate FPNc.

            * `gui_select_correction_axis`: int (0/1), 'Both'
                This parameter is used only when FPN_flag_GUI is True, otherwise it will be ignored.

            * `object_update_progressBar`: object
                Object that updates the progress bar in GUI.
        """
        super(DifferentialRollingAverage, self).__init__()

        self.cpu = CPUConfigurations()

        self.flag_GUI = flag_GUI
        self.mode_FPN = mode_FPN
        self.FPN_flag_run = FPN_flag_GUI
        self.DRA_flag_run = True
        self.select_correction_axis_run = gui_select_correction_axis

        self.size_A_diff = (video.shape[0] - 2 * batchSize, video.shape[1], video.shape[2])

        self.video = video.astype(np.float64)
        self.batchSize = batchSize

        self.moving_avg = np.empty_like(self.video)
        self.output_diff = np.empty(self.size_A_diff)

        self.output_batch_1 = np.empty(self.size_A_diff)
        self.fft_output_len = self.video.shape[0] - self.batchSize

        self.FPNc_video = None
        self.flag_thread_FPNc = True
        self.object_update_progressBar = object_update_progressBar
        self.threadpool = QThreadPool()
        self.signals = WorkerSignals()

    @Slot()
    def check_DRA_finish(self, flag):
        self.DRA_flag_run = flag

    @Slot()
    def run(self, *args, **kwargs):
        self.signals.DRA_complete_signal.connect(self.check_DRA_finish)
        video_DRA = self.differential_rolling(FPN_flag=self.FPN_flag_run, select_correction_axis=self.select_correction_axis_run,
                                              FFT_flag=False, inter_flag_parallel_active=True)

        self.signals.result_final.emit(video_DRA)

    @Slot()
    def thread_FPNc_complete(self):
        self.object_update_progressBar.setRange(self.p_max)
        self.object_update_progressBar.setLabel('')
        print("THREAD FPNc COMPLETE!")

    @Slot()
    def result_tread_FPNc(self, result):
        self.FPNc_video = result
        self.flag_thread_FPNc = False
        print("FPNc video update!")

    @Slot()
    def startProgressBarFPNc(self, instance, **kwargs):
        self.thread_FPNc = instance(**kwargs)
        self.thread_FPNc.signals.updateProgress_FPNC.connect(self.object_update_progressBar.setProgress)
        self.thread_FPNc.signals.finished_FPNc.connect(self.thread_FPNc_complete)
        self.thread_FPNc.signals.result_FPNc.connect(self.result_tread_FPNc)

        self.threadpool.start(self.thread_FPNc)

    def differential_rolling(self, FPN_flag=False, select_correction_axis=1, FFT_flag=False,
                             inter_flag_parallel_active=True,
                             max_iterations=10, FFT_widith=1):
        """
        To use DRA, you'll need to call the "differential rolling" process.

        Parameters
        ----------
        FPN_flag: bool
            This flag activates the fixed pattern noise correction function in case define as true.

        select_correction_axis: int (0/1), 'Both'
            This parameter is used only when FPN_flag is True, otherwise it will be ignored.

            * 0: FPN will be applied row-wise.
            * 1: FPN will be applied column-wise.
            * 'Both': FPN will be applied on two axis.

        FFT_flag: bool
            In case it is True, DRA will be performed in parallel to improve the time performance.

        inter_flag_parallel_active: bool
            This flag actives/inactives parallel computation of wFPNc.

        max_iterations: int
            This parameter is used when fFPT is selected that defines the total number of filtering iterations.

        FFT_widith: int
            This parameter is used when fFPT is selected that defines the frequency mask's width.

        Returns
        -------
        output: NDArray
            Returns DRA video.

        gainMap1D_: NDArray
            Returns projection on each frame based on the correction axis
        """
        if FPN_flag and self.mode_FPN == 'cpFPN':
            print("\n--- start DRA + cpFPN_axis: " + str(select_correction_axis) + "---")
        elif FPN_flag and self.mode_FPN == 'wFPN':
            print("\n--- start DRA + wFPN_axis: " + str(select_correction_axis) + "---")
        elif FPN_flag and self.mode_FPN == 'fFPN':
            print("\n--- start DRA + fFPN_axis: " + str(select_correction_axis) + "---")
        elif FPN_flag and self.mode_FPN == 'mFPN':
            print("\n--- start DRA + mFPN_axis: " + str(select_correction_axis) + "---")
        else:
            print("\n--- start DRA ---")

        self._apply_moving_average(FFT_flag)

        if FPN_flag:
            if self.mode_FPN == 'mFPN' or self.mode_FPN == 'cpFPN':
                if self.flag_GUI is True:

                    self.DRA_flag_run = True

                    while self.DRA_flag_run:
                        QtCore.QCoreApplication.processEvents()

                    time.sleep(2)

                    video = self._FPNc_GUI_axis(self.output_diff, select_correction_axis, self.mode_FPN)

                    time.sleep(2)


                else:
                    out_diff = self.output_diff
                    video = self._FPNc_axis(out_diff, select_correction_axis, self.mode_FPN, inter_flag_parallel_active)
            else:
                video = self.output_diff
        else:
            video = self.output_diff

        output = numba_divide(video, self.output_batch_1)

        if FPN_flag:

            if self.mode_FPN == 'wFPN' or self.mode_FPN == 'fFPN':

                output = self._FPNc_axis(output, select_correction_axis, self.mode_FPN, inter_flag_parallel_active,
                                         max_iterations, FFT_widith)

        if select_correction_axis == 'Both':
            gainMap1D_1 = np.mean(output, axis=1)
            gainMap1D_2 = np.mean(output, axis=2)
            gainMap1D_ = [gainMap1D_1, gainMap1D_2]

        elif select_correction_axis is not None:
            gainMap1D_ = np.mean(output, axis=select_correction_axis + 1)

        else:
            gainMap1D_ = None

        return output, gainMap1D_

    def _apply_moving_average(self, FFT_flag):
        if FFT_flag:
            self.movingAvg_FFT_based()
        else:
            if self.flag_GUI:
                self.DRA_flag_run = True
                self.object_update_progressBar.setLabel('DRA')
            self.temporal_moving_average()

    def temporal_moving_average(self):

        batch_1 = np.sum(self.video[0:self.batchSize, :, :], axis=0)
        batch_2 = np.sum(self.video[self.batchSize:2 * self.batchSize, :, :], axis=0)

        batch_1_ = np.divide(batch_1, self.batchSize)
        batch_2_ = np.divide(batch_2, self.batchSize)

        self.output_diff[0, :, :] = batch_2_ - batch_1_
        self.output_batch_1[0, :, :] = batch_1_

        for i_ in tqdm(range(1, self.video.shape[0] - 2 * self.batchSize)):
            if self.flag_GUI is True:
                self.signals.updateProgress_DRA.emit(i_)
            batch_1 = batch_1 - self.video[i_ - 1, :, :] + self.video[self.batchSize + i_ - 1, :, :]
            batch_2 = batch_2 - self.video[self.batchSize + i_ - 1, :, :] + self.video[(2 * self.batchSize) + i_ - 1, :, :]
            batch_1_ = np.divide(batch_1, self.batchSize)
            batch_2_ = np.divide(batch_2, self.batchSize)

            self.output_diff[i_, :, :] = (batch_2_ - batch_1_)
            self.output_batch_1[i_, :, :] = batch_1_

        if self.flag_GUI is True:
            self.signals.finished_DRA.emit()

    def numba_temporal_moving_average(self):

        batch_1 = numba_sum(self.video[0:self.batchSize, :, :], axis=0)
        batch_2 = numba_sum(self.video[self.batchSize:2 * self.batchSize, :, :], axis=0)

        batch_1_ = numba_divide(batch_1, self.batchSize)
        batch_2_ = numba_divide(batch_2, self.batchSize)

        self.output_diff[0, :, :] = numba_diff(batch_2_, batch_1_)
        self.output_batch_1[0, :, :] = batch_1_

        for i_ in tqdm(range(1, self.video.shape[0] - 2 * self.batchSize)):
            if self.flag_GUI is True:
                self.signals.updateProgress_DRA.emit(i_)
            batch_1 = batch_1 - self.video[i_ - 1, :, :] + self.video[self.batchSize + i_ - 1, :, :]
            batch_2 = batch_2 - self.video[self.batchSize + i_ - 1, :, :] + self.video[(2 * self.batchSize) + i_ - 1, :, :]

            batch_1_ = numba_divide(batch_1, self.batchSize)
            batch_2_ = numba_divide(batch_2, self.batchSize)

            self.output_diff[0, :, :] = numba_diff(batch_2_, batch_1_)
            self.output_batch_1[0, :, :] = batch_1_

        if self.flag_GUI is True:
            self.signals.finished_DRA.emit()

    def movingAvg_FFT_based(self):

        Parallel(n_jobs=self.cpu.n_jobs, backend='threading')(delayed(self.uniform_filter1d_kernel)(r_, c_)
                            for r_ in range(self.video.shape[1])
                            for c_ in range(self.video.shape[2]))

        self.output_diff = self.moving_avg[self.batchSize:self.fft_output_len, ...] - self.moving_avg[0:self.fft_output_len-self.batchSize, ...]

        self.output_batch_1 = self.moving_avg[0:self.fft_output_len - self.batchSize, ...]

    def uniform_filter1d_kernel(self, r_, c_):
        self.moving_avg[:, r_, c_] = uniform_filter1d(self.video[:, r_, c_], size=self.batchSize,
                                                      mode='constant', cval=0.0,
                                                      origin=-(self.batchSize // 2), axis=0)

    def _FPNc_axis(self, video, select_correction_axis, mode_FPN, inter_flag_parallel_active=True, max_iterations=10,
                   FFT_widith=1):
        if mode_FPN == 'cpFPN' or mode_FPN == 'mFPN':
            if select_correction_axis == 'Both':
                vid_FPNc_axis_1 = self._FPN(video, 1, mode_FPN, inter_flag_parallel_active, max_iterations, FFT_widith)
                vid_FPNc = self._FPN(vid_FPNc_axis_1, 0, mode_FPN, inter_flag_parallel_active, max_iterations,
                                     FFT_widith)

            else:
                vid_FPNc = self._FPN(video, select_correction_axis, mode_FPN, inter_flag_parallel_active,
                                     max_iterations, FFT_widith)
        else:
            vid_FPNc = self._FPN(video, select_correction_axis, mode_FPN, inter_flag_parallel_active, max_iterations,
                                 FFT_widith)

        return vid_FPNc

    def _FPN(self, video, select_correction_axis, mode_FPN, inter_flag_parallel_active=True, max_iterations=10,
             FFT_widith=1):
        if mode_FPN == 'mFPN':
            noise_c = FPNc.MedianProjectionFPNc(video=video,
                                                select_correction_axis=select_correction_axis,
                                                flag_GUI=self.flag_GUI)
            video_FPNc = noise_c.mFPNc(select_correction_axis)

        elif mode_FPN == 'cpFPN':
            noise_c = FPNc.ColumnProjectionFPNc(video=video,
                                                select_correction_axis=select_correction_axis,
                                                flag_GUI=self.flag_GUI)
            video_FPNc = noise_c.cpFPNc(select_correction_axis)

        elif mode_FPN == 'wFPN':
            if select_correction_axis == 0:
                wFPN = FPNc.FrequencyFPNc(video, inter_flag_parallel_active=inter_flag_parallel_active)
                video_FPNc = wFPN.update_wFPN(direction='Horizontal')

            elif select_correction_axis == 1:
                wFPN = FPNc.FrequencyFPNc(video, inter_flag_parallel_active=inter_flag_parallel_active)
                video_FPNc = wFPN.update_wFPN(direction='Vertical')

            elif select_correction_axis == 'Both':
                wFPN = FPNc.FrequencyFPNc(video, inter_flag_parallel_active=inter_flag_parallel_active)
                output = wFPN.update_wFPN(direction='Horizontal')

                wFPN = FPNc.FrequencyFPNc(output, inter_flag_parallel_active=inter_flag_parallel_active)
                video_FPNc = wFPN.update_wFPN(direction='Vertical')

        elif mode_FPN == 'fFPN':
            if select_correction_axis == 0:
                wf_FPN = FPNc.FrequencyFPNc(video, inter_flag_parallel_active=inter_flag_parallel_active)
                video_FPNc = wf_FPN.update_fFPN(direction='Horizontal', max_iterations=max_iterations, width=FFT_widith)

            elif select_correction_axis == 1:
                wf_FPN = FPNc.FrequencyFPNc(video, inter_flag_parallel_active=inter_flag_parallel_active)
                video_FPNc = wf_FPN.update_fFPN(direction='Vertical', max_iterations=max_iterations, width=FFT_widith)

            elif select_correction_axis == 'Both':
                wf_FPN = FPNc.FrequencyFPNc(video, inter_flag_parallel_active=inter_flag_parallel_active)
                output = wf_FPN.update_fFPN(direction='Horizontal', max_iterations=max_iterations, width=FFT_widith)

                wf_FPN = FPNc.FrequencyFPNc(output, inter_flag_parallel_active=inter_flag_parallel_active)
                video_FPNc = wf_FPN.update_fFPN(direction='Vertical', max_iterations=max_iterations, width=FFT_widith)

        return video_FPNc

    def _FPNc_GUI_axis(self, video, select_correction_axis, mode_FPN):

        if select_correction_axis == 'Both':

            self.flag_thread_FPNc = True
            vid_FPNc_axis_1 = self._FPN_GUI(video, 1, mode_FPN)

            while self.flag_thread_FPNc:
                QtCore.QCoreApplication.processEvents()

            time.sleep(2)

            self.flag_thread_FPNc = True
            vid_FPNc = self._FPN_GUI(vid_FPNc_axis_1, 0, mode_FPN)

            while self.flag_thread_FPNc:
                QtCore.QCoreApplication.processEvents()

            time.sleep(2)

        else:
            self.flag_thread_FPNc = True
            vid_FPNc = self._FPN_GUI(video, select_correction_axis, mode_FPN)

        return vid_FPNc

    def _FPN_GUI(self, video, select_correction_axis, mode_FPN):

        if mode_FPN == 'mFPN':
            d_arg = {'video': video, 'flag_GUI': True,
                     'instance': FPNc.MedianProjectionFPNc,
                     'select_correction_axis': select_correction_axis}
        elif mode_FPN == 'cpFPN':
            d_arg = {'video': video, 'flag_GUI': True,
                     'instance': FPNc.ColumnProjectionFPNc,
                     'select_correction_axis': select_correction_axis}

        self.p_max = video.shape[0] - (2 * self.batchSize) - 1
        self.object_update_progressBar.setRange(self.p_max)
        self.object_update_progressBar.setProgress(0)
        self.object_update_progressBar.setLabel('FPNc')

        self.flag_thread_FPNc = True
        self.startProgressBarFPNc(**d_arg)

        while self.flag_thread_FPNc:
            QtCore.QCoreApplication.processEvents()

        time.sleep(2)

        return self.FPNc_video














from piscat.GUI.InputOutput import Reading
from PySide6 import QtCore, QtWidgets


class BgCorrection_GUI(QtWidgets.QWidget):

    output_setting_Tab_bgCorrection = QtCore.Signal(object)
    update_tab_index = QtCore.Signal(int)

    def __init__(self, parent=None):
        super(BgCorrection_GUI, self).__init__(parent)
        self.different_views = {}

        self.setting_bg_correction = {}

        self.flag_remove_box = True
        self.flag_update_class = True
        self.flag_hotPixels = False

        self.empty_value_box_flag = False
        self.flag_no_warning = True

        self.transform_algorithm = None
        self.n_comp = None
        self.n_iter = None
        self.alpha = None
        self.random_select = None
        self.non_zero_coeff = None
        self.patch_size = None
        self.strides = None

        self.resize(600, 600)
        self.setWindowTitle('Background Correction')

        self.combo_bg_filter = QtWidgets.QComboBox(self)
        self.combo_bg_filter.addItem("-Select the background correction methods-")
        self.combo_bg_filter.addItem("Differential rolling average")
        self.combo_bg_filter.addItem("Spatial median Filter")
        self.combo_bg_filter.addItem("Spatial gaussian Filter")
        self.combo_bg_filter.addItem("Background correction temporal median")
        self.combo_bg_filter.addItem("Flat field (Gaussian filter)")
        self.combo_bg_filter.addItem("Flat field (mean background subtraction)")
        self.combo_bg_filter.currentIndexChanged.connect(self.layout_based_on_select_bg)

        self.btn = QtWidgets.QPushButton("Next")
        self.btn.setAutoDefault(False)
        self.btn.clicked.connect(self.do_update)
        self.btn.setFixedWidth(70)
        self.btn.setFixedHeight(20)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.combo_bg_filter, 0, 0)
        self.grid.addWidget(self.createFirstExclusiveGroup(), 1, 0)
        self.grid.addWidget(self.btn, 2, 0)

        self.setLayout(self.grid)

    def __del__(self):
        print('Destructor called, Employee deleted.')

    def createFirstExclusiveGroup(self):

        self.groupBox_update = QtWidgets.QGroupBox("Background correction Setting:")
        self.grid1 = QtWidgets.QGridLayout()

        self.groupBox_update.setLayout(self.grid1)
        return self.groupBox_update

    def massage_filled(self):
        self.flag_no_warning = False
        self.msg_box1 = QtWidgets.QMessageBox()
        self.msg_box1.setWindowTitle("Warning!")
        self.msg_box1.setText("Please filled all parameters!")
        self.msg_box1.exec_()

    @QtCore.Slot()
    def do_update(self):
        if self.btn.clicked:
            if self.combo_bg_filter.currentText() == "Differential rolling average":
                self.batch_size = self.le_batchSize.text()

                if self.groupBox_FPNc.isChecked():
                    self.flag_FPN = True
                    if self.radio_axis_1.isChecked():
                        self.axis = 0
                    elif self.radio_axis_2.isChecked():
                        self.axis = 1
                    elif self.radio_axis_3.isChecked():
                        self.axis = 'Both'

                    if self.radio_cpFPN_mode.isChecked():
                        self.mode_FPN = 'cpFPN'
                    elif self.radio_wFPN_mode.isChecked():
                        self.mode_FPN = 'wFPN'
                    elif self.radio_FFT2D_FPN_mode.isChecked():
                        self.mode_FPN = 'fFPN'
                    elif self.radio_median_FPN_mode.isChecked():
                        self.mode_FPN = 'mFPN'
                else:
                    self.flag_FPN = False
                    self.axis = 0
                    self.mode_FPN = 'fFPN'

                if self.checkbox_power_normalization.isChecked():
                    self.flag_power_normalization = True
                else:
                    self.flag_power_normalization = False

                if self.checkbox_hotPixels.isChecked():
                    self.flag_hotPixels = True
                else:
                    self.flag_hotPixels = False

                if self.batch_size != '':
                    self.batch_size = int(self.batch_size)

                    self.setting_bg_correction['type_BGc'] = 'DRA'
                    self.setting_bg_correction['mode_FPN'] = self.mode_FPN
                    self.setting_bg_correction['Batch_size (frames)'] = self.batch_size
                    self.setting_bg_correction['Power_Normalization'] = self.flag_power_normalization
                    self.setting_bg_correction['FPNc'] = self.flag_FPN
                    self.setting_bg_correction['FPNc_axis'] = self.axis
                    self.setting_bg_correction['filter_hotPixels'] = self.flag_hotPixels

                    self.output_setting_Tab_bgCorrection.emit(self.setting_bg_correction)

                else:
                    self.msg_box = QtWidgets.QMessageBox()
                    self.msg_box.setWindowTitle("Warning!")
                    self.msg_box.setText("Please insert batch size!")
                    self.msg_box.exec_()

            elif self.combo_bg_filter.currentText() == "Spatial median Filter":
                median_size = self.line_edit1.text()
                if median_size != '':
                    print("\nMedian is Applying--->", end='')
                    median_size = int(self.line_edit1.text())

                    self.setting_bg_correction['type_BGc'] = 'Spatial median Filter'
                    self.setting_bg_correction['Spatial_median_kernel_size'] = median_size
                    self.output_setting_Tab_bgCorrection.emit(self.setting_bg_correction)
                    print("Done!")
                else:
                    self.massage_filled()

            elif self.combo_bg_filter.currentText() == "Spatial gaussian Filter":
                gaussian_sigma = self.line_edit2.text()
                if gaussian_sigma != '':
                    print("\nGaussian is applying--->", end='')
                    gaussian_sigma = float(self.line_edit2.text())
                    self.setting_bg_correction['type_BGc'] = 'Spatial gaussian Filter'
                    self.setting_bg_correction['Spatial_gaussian_sigma'] = gaussian_sigma
                    self.output_setting_Tab_bgCorrection.emit(self.setting_bg_correction)
                    print("Done!")
                else:
                    self.massage_filled()

            elif self.combo_bg_filter.currentText() == "Background correction temporal median":
                print("\nBackground correction temporal median--->", end='')
                self.setting_bg_correction['type_BGc'] = 'Background correction temporal median'
                self.output_setting_Tab_bgCorrection.emit(self.setting_bg_correction)
                print("Done!")

            elif self.combo_bg_filter.currentText() == "Flat field (Gaussian filter)":
                sigma = self.line_edit5.text()
                if sigma != '':
                    print("\nBackground correction Flat Field--->", end='')
                    sigma = int(self.line_edit5.text())
                    self.setting_bg_correction['type_BGc'] = 'Flat field (Gaussian filter)'
                    self.setting_bg_correction['Flat_field_sigma'] = sigma
                    self.output_setting_Tab_bgCorrection.emit(self.setting_bg_correction)
                    print("Done!")
                else:
                    self.massage_filled()

            elif self.combo_bg_filter.currentText() == "Flat field (mean background subtraction)":
                if self.original_video_bg is not None:
                    print("\nBackground correction Flat Field back ground--->", end='')
                    self.setting_bg_correction['type_BGc'] = 'Flat field (mean background subtraction)"'
                    self.output_setting_Tab_bgCorrection.emit(self.setting_bg_correction)
                    print("Done!")

                else:
                    self.massage_filled()
            else:
                self.msg_box = QtWidgets.QMessageBox()
                self.msg_box.setWindowTitle("Warning!")
                self.msg_box.setText("Continuing without using any filter!")
                self.msg_box.exec_()
                self.setting_bg_correction['type_BGc'] = 'RAW'
                self.output_setting_Tab_bgCorrection.emit(self.setting_bg_correction)

            self.update_tab_index.emit(2)

    def on_select(self):
        self.transform_algorithm = self.combo.currentText()
        self.create_como_values()

    def axis_selection(self):
        if self.radio_cpFPN_mode.isChecked():
            self.radio_axis_1.setEnabled(True)
            self.radio_axis_2.setEnabled(True)
            self.radio_axis_3.setEnabled(True)

        elif self.radio_wFPN_mode.isChecked():
            self.radio_axis_1.setEnabled(True)
            self.radio_axis_2.setEnabled(True)
            self.radio_axis_3.setEnabled(True)

        elif self.radio_median_FPN_mode.isChecked():
            self.radio_axis_1.setEnabled(True)
            self.radio_axis_2.setEnabled(True)
            self.radio_axis_3.setEnabled(True)

        elif self.radio_FFT2D_FPN_mode.isChecked():
            self.radio_axis_1.setEnabled(True)
            self.radio_axis_2.setEnabled(True)
            self.radio_axis_3.setEnabled(True)

    def layout_based_on_select_bg(self):
        if self.combo_bg_filter.currentText() == "Differential rolling average":
            while self.flag_remove_box:
                self.remove_widgets()
            self.flag_remove_box = True
            self.groupBatchSize = QtWidgets.QGroupBox("Batch size:")

            self.le_batchSize = QtWidgets.QLineEdit()
            self.le_batchSize.setPlaceholderText('radius')
            self.le_batchSize_label = QtWidgets.QLabel("Batch size (frames):")
            self.le_batchSize.setFixedWidth(50)

            self.checkbox_power_normalization = QtWidgets.QCheckBox("Laser power normalization", self)
            self.checkbox_hotPixels = QtWidgets.QCheckBox("Hot pixels correction", self)

            grid_batch = QtWidgets.QGridLayout()
            grid_batch.addWidget(self.le_batchSize_label, 0, 0)
            grid_batch.addWidget(self.le_batchSize, 0, 1)
            grid_batch.addWidget(self.checkbox_power_normalization, 1, 0)
            grid_batch.addWidget(self.checkbox_hotPixels, 2, 0)

            self.groupBatchSize.setLayout(grid_batch)

            self.groupBox_FPNc = QtWidgets.QGroupBox("FPNc:")
            self.groupBox_FPNc.setCheckable(True)
            self.groupBox_FPNc.setChecked(False)

            self.FPN_mode_group = QtWidgets.QButtonGroup()
            self.radio_wFPN_mode = QtWidgets.QRadioButton("Wavelet FPNc")
            self.radio_cpFPN_mode = QtWidgets.QRadioButton("Column_projection FPNc")
            self.radio_median_FPN_mode = QtWidgets.QRadioButton("Median FPNc")
            self.radio_FFT2D_FPN_mode = QtWidgets.QRadioButton("FFT2D FPNc")

            self.FPN_mode_group.addButton(self.radio_wFPN_mode)
            self.FPN_mode_group.addButton(self.radio_cpFPN_mode)
            self.FPN_mode_group.addButton(self.radio_FFT2D_FPN_mode)
            self.FPN_mode_group.addButton(self.radio_median_FPN_mode)

            self.radio_wFPN_mode.toggled.connect(self.axis_selection)
            self.radio_cpFPN_mode.toggled.connect(self.axis_selection)
            self.radio_FFT2D_FPN_mode.toggled.connect(self.axis_selection)
            self.radio_median_FPN_mode.toggled.connect(self.axis_selection)

            self.axis_group = QtWidgets.QButtonGroup()
            self.radio_axis_1 = QtWidgets.QRadioButton("FPNc in axis 0")
            self.radio_axis_2 = QtWidgets.QRadioButton("FPNc in axis 1")
            self.radio_axis_3 = QtWidgets.QRadioButton("FPNc in Both axis")
            self.radio_axis_2.setChecked(True)
            self.radio_cpFPN_mode.setChecked(True)

            self.axis_group.addButton(self.radio_axis_1)
            self.axis_group.addButton(self.radio_axis_2)
            self.axis_group.addButton(self.radio_axis_3)

            grid = QtWidgets.QGridLayout()
            grid.addWidget(self.radio_cpFPN_mode, 0, 0)
            grid.addWidget(self.radio_median_FPN_mode, 1, 0)
            grid.addWidget(self.radio_wFPN_mode, 2, 0)
            grid.addWidget(self.radio_FFT2D_FPN_mode, 3, 0)

            grid.addWidget(self.radio_axis_1, 0, 1)
            grid.addWidget(self.radio_axis_2, 1, 1)
            grid.addWidget(self.radio_axis_3, 2, 1)

            self.groupBox_FPNc.setLayout(grid)

            self.grid1.addWidget(self.groupBatchSize, 1, 0)
            self.grid1.addWidget(self.groupBox_FPNc, 2, 0)

        elif self.combo_bg_filter.currentText() == "Spatial median Filter":

            while self.flag_remove_box:
                self.remove_widgets()

            self.flag_remove_box = True

            self.line_edit1 = QtWidgets.QLineEdit(self)
            self.line_edit1_label = QtWidgets.QLabel("Neighborhood size (px):")
            self.line_edit1.setFixedWidth(50)

            self.grid1.addWidget(self.line_edit1_label, 2, 0)
            self.grid1.addWidget(self.line_edit1, 2, 1)

        elif self.combo_bg_filter.currentText() == "Spatial gaussian Filter":

            while self.flag_remove_box:
                self.remove_widgets()

            self.flag_remove_box = True

            self.line_edit2 = QtWidgets.QLineEdit(self)
            self.line_edit2_label = QtWidgets.QLabel("Sigma (px):")
            self.line_edit2.setFixedWidth(50)

            self.grid1.addWidget(self.line_edit2_label, 2, 0)
            self.grid1.addWidget(self.line_edit2, 2, 1)

        elif self.combo_bg_filter.currentText() == "Background correction temporal median":

            while self.flag_remove_box:
                self.remove_widgets()

            self.flag_remove_box = True

        elif self.combo_bg_filter.currentText() == "Flat field (Gaussian filter)":

            while self.flag_remove_box:
                self.remove_widgets()

            self.flag_remove_box = True

            self.line_edit5 = QtWidgets.QLineEdit(self)
            self.line_edit5_label = QtWidgets.QLabel("Sigma (px):")
            self.line_edit5.setFixedWidth(50)

            self.grid1.addWidget(self.line_edit5_label, 2, 0)
            self.grid1.addWidget(self.line_edit5, 2, 1)

        elif self.combo_bg_filter.currentText() == "Flat field (mean background subtraction)":

            while self.flag_remove_box:
                self.remove_widgets()

            self.flag_remove_box = True

            self.btn_read_file = QtWidgets.QPushButton("load")
            self.btn_read_file.setAutoDefault(False)
            self.btn_read_file.clicked.connect(self.load_video)
            self.btn_read_file.setFixedHeight(20)
            self.btn_read_file.setFixedWidth(50)

            self.grid1.addWidget(self.btn_read_file, 1, 0)

    def remove_widgets(self):
        for i_ in range(0, 7, 1):
            for j_ in range(0, 7, 1):
                if i_ == 0 and j_ == 0:
                    pass
                else:
                    layout = self.grid1.itemAtPosition(i_, j_)
                    if layout is not None:
                        layout.widget().deleteLater()
                        self.grid1.removeItem(layout)

        self.flag_remove_box = False

    def load_video(self):
        self.reading = Reading()
        self.reading.update_output.connect(self.updata_bg_video)
        self.reading.read_video()

    def closeEvent(self, event):
        QtCore.QCoreApplication.instance().quit()



from PySide6 import QtWidgets, QtCore, QtGui


class PlaySetting(QtWidgets.QWidget):
    procDone_Animation_panel = QtCore.Signal(object)

    def __init__(self, parent=None):
        super(PlaySetting, self).__init__(parent)

        # ok
        self.ok = QtWidgets.QPushButton("ok")
        self.ok.setAutoDefault(False)
        self.ok.clicked.connect(self.do_update)
        self.ok.setFixedWidth(100)

        self.le1 = QtWidgets.QLineEdit()
        self.le1.setPlaceholderText('step_size')
        self.le_1_label = QtWidgets.QLabel("Stride between frame:")
        self.le1.setFixedWidth(100)

        self.le2 = QtWidgets.QLineEdit()
        self.le2.setPlaceholderText('Time delay')
        self.le_2_label = QtWidgets.QLabel("Time delay (ms):")
        self.le2.setFixedWidth(100)

        self.le3 = QtWidgets.QLineEdit()
        self.le3.setPlaceholderText('fps')
        self.le_3_label = QtWidgets.QLabel("Speed:")
        self.le3.setFixedWidth(100)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.le_1_label, 0, 0)
        self.grid.addWidget(self.le1, 0, 1)
        self.grid.addWidget(self.le_2_label, 1, 0)
        self.grid.addWidget(self.le2, 1, 1)
        self.grid.addWidget(self.le_3_label, 2, 0)
        self.grid.addWidget(self.le3, 2, 1)
        self.grid.addWidget(self.ok, 3, 1)
        self.setLayout(self.grid)

        self.setWindowTitle("Animation panel")
        self.setLayout(self.grid)

    @QtCore.Slot()
    def do_update(self):

        try:
            self.step_size = int(self.le1.text())
        except:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Incorrect input_video for step size! \nThe default value will be selected")
            self.msg_box.exec_()
            self.step_size = 1

        try:
            self.time_delay = float(self.le2.text())
        except:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Incorrect input_video for time delay! \nThe default value will be selected")
            self.msg_box.exec_()
            self.time_delay = 0.1

        try:
            self.fps = float(self.le3.text())
        except:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Incorrect input_video for Speed! \nThe default value will be selected")
            self.msg_box.exec_()
            self.fps = 10.0

        self.procDone_Animation_panel.emit([self.step_size, self.time_delay, self.fps])
        self.close()


from __future__ import print_function

import numpy as np
import pandas as pd

from joblib import Parallel, delayed
from scipy import signal
from tqdm.autonotebook import tqdm

from piscat.InputOutput.cpu_configurations import CPUConfigurations
from piscat.Preproccessing import patch_genrator


class RadialCenter:

    def __init__(self):

        """
        The RadialCenter localization algorithm is implemented in Python. 

        References
        ----------
        Parthasarathy, R. Rapid, accurate particle tracking by calculation of radial symmetry centers. Nat Methods 9, 724–726 (2012).
        https://doi.org/10.1038/nmeth.2071
        """

        self.cpu = CPUConfigurations()

        self.patch_gen = None
        self.patch = None
        self.video_shape = None
        self.probility_map = None

    def psf_center_all_frames(self, video):
        p_center_all = []

        if self.cpu.parallel_active is True:
            print('\n---Center localization with parallel loop---')

            result = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(
                delayed(self.radialcenter)(video[i_, :, :]) for i_ in range(video.shape[0]))
            columns_name = ['x', 'y', 'sigma']
            df_psf = pd.DataFrame(result, columns=columns_name)
        else:
            print('\n---Center localization without parallel loop---')
            for i_ in range(video.shape[0]):
                xc, yc, sigma = self.radialcenter(video[i_, :, :])
                p_center_all.append([xc, yc, sigma])

            columns_name = ['x', 'y', 'sigma']
            df_psf = pd.DataFrame(p_center_all, columns=columns_name)
        return df_psf

    def patch_genrator(self, original_video, patch_size=16, strides=4):
        self.video_shape = original_video.shape
        self.patch_gen = patch_genrator.ImagePatching()
        self.patch = self.patch_gen.split_weight_matrix(original_video, patch_size=patch_size, strides=strides)

    def radialcenter_probility_map(self):
        if self.patch is not None and self.patch_gen is not None and self.patch_gen is not None:
            patch_probility = []
            for p_ in tqdm(self.patch):
                tmp_ = np.zeros(p_.shape)
                self.parallel_active = False
                df_psf = self.psf_center_all_frames(p_)
                y_list = df_psf['y'].to_list()
                x_list = df_psf['y'].to_list()
                f_list = list(range(0, df_psf.shape[0], 1))

                y_list_ = [int(v_) for v_ in y_list]
                x_list_ = [int(v_) for v_ in x_list]

                try:
                    tmp_[f_list, y_list_, x_list_] = 1
                    patch_probility.append(tmp_)

                except:
                    patch_probility.append(tmp_)

            self.probility_map = self.patch_gen.reconstruction_weight_matrix(self.video_shape, new_patch=patch_probility)
            return self.probility_map
        else:
            print('you need to call patch_genrator method!')

    def radialcenter(self, Image):
        Image = Image.astype(np.float64)
        Nx = Image.shape[0]
        Ny = Image.shape[1]

        xm_onerow = np.arange(-((Nx-1)/2.0)+0.5, ((Nx-1)/2.0)+0.5, 1, dtype=np.int)
        xm = np.broadcast_to(xm_onerow, (Nx-1, Ny-1))

        ym_onecol = np.arange(-((Ny-1)/2.0)+0.5, ((Ny-1)/2.0)+0.5,  dtype=np.int)
        ym = np.transpose(np.broadcast_to(ym_onecol, (Nx-1, Ny-1)))

        dIdu = Image[0:Ny-1, 1:Nx] - Image[1:Ny, 0:Nx-1]
        dIdv = Image[0:Ny-1, 0:Nx-1] - Image[1:Ny, 1:Nx]

        h = (1/9) * np.ones((3, 3))
        fdu = signal.convolve2d(dIdu, h, mode='same')
        fdv = signal.convolve2d(dIdv, h, mode='same')
        dImag2 = np.multiply(fdu, fdu) + np.multiply(fdv, fdv)

        m = -np.divide((fdv + fdu), (fdu - fdv))

        NNanm = np.sum(np.isnan(m))
        if NNanm > 0:
            unsmoothm = np.divide((dIdv + dIdu), (dIdu - dIdv))
            m[np.isnan(m)] = unsmoothm[np.isnan(m)]

        NNanm = np.sum(np.isnan(m))
        if NNanm > 0:
            m[np.isnan(m)] = 0

        try:
            m[np.isinf(m)] = 10 * np.max(m[~np.isinf(m)])
        except:
            m = np.divide((dIdv + dIdu), (dIdu-dIdv))

        b = ym - np.multiply(m, xm)
        sdI2 = np.sum(dImag2)
        xcentroid = np.sum(np.sum(np.multiply(dImag2, xm))) / sdI2
        ycentroid = np.sum(np.sum(np.multiply(dImag2, ym))) / sdI2
        w = np.divide(dImag2, np.sqrt(np.multiply((xm - xcentroid), (xm - xcentroid)) + np.multiply((ym - ycentroid), (ym - ycentroid))))

        xc, yc = self.lsradialcenterfit(m, b, w)

        xc = xc + (Nx + 1) / 2.0
        yc = yc + (Ny + 1) / 2.0

        Isub = Image - np.min(Image)
        x = np.linspace(0, Nx - 1, Nx, dtype=np.int)
        y = np.linspace(0, Ny - 1, Ny, dtype=np.int)
        px, py = np.meshgrid(x, y)
        xoffset = px - xc
        yoffset = py - yc
        r2 = np.multiply(xoffset, xoffset) + np.multiply(yoffset, yoffset)
        sigma = np.sqrt(np.sum(np.sum(np.multiply(Isub, r2))) / np.sum(Isub)) / 2

        return xc, yc, sigma

    def lsradialcenterfit(self, m, b, w):

        wm2p1 = np.divide(w, np.multiply(m, m) + 1)
        sw = np.sum(wm2p1)
        smmw = np.sum(np.multiply(m, np.multiply(m, wm2p1)))
        smw = np.sum(np.multiply(m, wm2p1))
        smbw = np.sum(np.multiply(m, np.multiply(b, wm2p1)))
        sbw = np.sum(np.multiply(b, wm2p1))
        det = smw * smw - smmw * sw
        xc = (smbw * sw - smw * sbw) / det
        yc = (smbw * smw - smmw * sbw) / det
        return xc, yc



# from GUI_RUN.Main import PiSCAT_GUI
from PySide6 import QtCore

class ProgressBar():

    def __init__(self, object):
        super(ProgressBar, self).__init__()
        self.progressBar = object

    @QtCore.Slot(int)
    def setProgress(self, val):
        self.progressBar.setValue(val)

    def setRange(self, max):
        self.progressBar.setValue(0)
        self.progressBar.setMaximum(max)

    def setLabel(self, str):
        self.progressBar.setFormat(str)



import numpy as np
import cmath


def L_theta(L, theta , p, ci, z, z_p):

    ni2sin2theta = p.ni ** 2 * cmath.sin(theta) ** 2
    sroot = cmath.sqrt(p.ns ** 2 - ni2sin2theta)
    groot = cmath.sqrt(p.ng ** 2 - ni2sin2theta)
    g0root = cmath.sqrt(p.ng0 ** 2 - ni2sin2theta)
    i0root = cmath.sqrt(p.ni0 ** 2 - ni2sin2theta)
    L[0] = p.ni * (ci - z) * cmath.cos(theta) + z_p * sroot + p.tg * groot - p.tg0 * g0root - p.ti0 * i0root
    L[1] = p.ni * cmath.sin(theta) * (
                z - ci + p.ni * cmath.cos(theta) * (p.tg0 / g0root + p.ti0 / i0root - p.tg / groot - z_p / sroot))

    return L



from PySide6.QtWidgets import QWidget, QApplication, QLabel
from PySide6.QtCore import QRect, Qt, QLine
from PySide6.QtGui import QImage, QPixmap, QPainter, QPen, QGuiApplication

from piscat.Localization.directional_intensity import DirectionalIntensity


class Line_Annotated(QLabel):
    x0 = 0
    y0 = 0
    x1 = 0
    y1 = 0
    flag = False

    def mousePressEvent(self, event):
        self.flag = True
        self.x0 = event.x()
        self.y0 = event.y()

    def mouseReleaseEvent(self, event):
        self.flag = False
        # self.read_pixel_valus()

    def mouseMoveEvent(self, event):
        if self.flag:
            self.x1 = event.x()
            self.y1 = event.y()
            self.update()

    def paintEvent(self, event):
        super().paintEvent(event)
        rect = QLine(self.x0, self.y0, self.x1, self.y1)

        painter = QPainter(self)
        painter.setPen(QPen(Qt.red, 2, Qt.SolidLine))
        painter.drawLine(rect)

    def read_pixel_valus(self):
        DI = DirectionalIntensity()
        radial_index = DI.interpolate_pixels_along_line(x0=int(self.x0), y0=int(self.y0), x1=int(self.x1),
                                                        y1=int(self.y1))

        for idx_ in radial_index:
            self.ori_X = int((self.orginal_video.shape[1] / self.pixmap_video.size().toTuple()[0]) * idx_[0])
            self.ori_Y = int((self.orginal_video.shape[2] / self.pixmap_video.size().toTuple()[1]) * idx_[1])



from __future__ import print_function

import numpy as np
import pywt

from tqdm.autonotebook import tqdm
from joblib import Parallel, delayed
from PySide6.QtCore import *
from tqdm.notebook import tqdm
from skimage.morphology import rectangle
from skimage.filters import rank
from skimage import filters
from piscat.InputOutput.cpu_configurations import CPUConfigurations
from piscat.Preproccessing.filtering import GuidedFilter
from piscat.Visualization.print_colors import PrintColors


class WorkerSignals(QObject):

    updateProgress_FPNC = Signal(int)
    result_FPNc = Signal(object)
    finished_FPNc = Signal()


class MedianProjectionFPNc(QRunnable):

    def __init__(self, video, select_correction_axis, flag_GUI=False):

        """
        This class uses a heuristic procedure called Median Projection FPN (mFPN) to reduce fixed pattern noise (FPN).

        References
        ----------
        [1] Mirzaalian Dastjerdi, Houman, et al. "Optimized analysis for sensitive detection and analysis of single
        proteins via interferometric scattering microscopy." Journal of Physics D: Applied Physics (2021).
        (http://iopscience.iop.org/article/10.1088/1361-6463/ac2f68)

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        select_correction_axis: int (0/1)
            * `0`: FPN will be applied row-wise.
            * `1`: FPN will be applied column-wise.
        """

        super(MedianProjectionFPNc, self).__init__()
        self.cpu = CPUConfigurations()
        self.video = video
        self.select_correction_axis = select_correction_axis
        self.flag_GUI = flag_GUI
        self.threadpool = QThreadPool()
        self.signals = WorkerSignals()

    @Slot()
    def run(self, *args, **kwargs):
        result = self.mFPNc(self.select_correction_axis)
        self.signals.result_FPNc.emit(result)

    def mFPNc(self, select_correction_axis):
        """
        Using the mPN approach on video.

        Parameters
        ----------
        select_correction_axis: int (0/1)

           * `0`: FPN will be applied row-wise.
           * `1`: FPN will be applied column-wise.

        Returns
        -------
        output: NDArray
            Video after using the mFPNc technique.
        """
        if False:
            print('\nFPN correction with parallel loop --->', end=" ")

            result = Parallel(n_jobs=self.cpu.n_jobs, backend='multiprocessing', verbose=self.cpu.verbose)(
                delayed(self.FPNc_kernel)(self.video[f_, :, :], select_correction_axis) for f_ in tqdm(range(0, self.video.shape[0])))
            output = np.asarray(result)

        else:

            print('\nmedian FPN correction without parallel loop --->', end=" ")
            output = np.zeros_like(self.video)

            for i_ in tqdm(range(0, self.video.shape[0])):

                if self.flag_GUI is True:
                    self.signals.updateProgress_FPNC.emit(i_)

                diff = self.video[i_, :, :]

                gainMap1D_median = np.median(diff, axis=select_correction_axis)

                if select_correction_axis == 1:
                    FPN2D_median = np.repeat(gainMap1D_median[:, np.newaxis], diff.shape[select_correction_axis],
                                             axis=select_correction_axis)
                elif select_correction_axis == 0:
                    FPN2D_median = np.repeat(gainMap1D_median[np.newaxis, :], diff.shape[select_correction_axis],
                                             axis=select_correction_axis)

                output[i_, :, :] = diff - FPN2D_median

        if self.flag_GUI is True:
            self.signals.finished_FPNc.emit()
        print('Done')

        return output

    def FPNc_kernel(self, diff, select_correction_axis):

        gainMap1D_median = np.median(diff, axis=select_correction_axis)

        if select_correction_axis == 1:
            FPN2D_median = np.repeat(gainMap1D_median[:, np.newaxis], diff.shape[select_correction_axis],
                                     axis=select_correction_axis)
        elif select_correction_axis == 0:
            FPN2D_median = np.repeat(gainMap1D_median[np.newaxis, :], diff.shape[select_correction_axis],
                                     axis=select_correction_axis)

        AD_absDev = np.abs(diff - FPN2D_median)
        MAD_1D_MedianAD = np.median(AD_absDev, axis=select_correction_axis)

        if select_correction_axis == 1:
            MAD_2D = np.repeat(MAD_1D_MedianAD[:, np.newaxis], diff.shape[select_correction_axis],
                               axis=select_correction_axis)
        elif select_correction_axis == 0:
            MAD_2D = np.repeat(MAD_1D_MedianAD[np.newaxis, :], diff.shape[select_correction_axis],
                               axis=select_correction_axis)

        diff_f1f2_nan = diff.copy()
        diff_f1f2_nan[MAD_2D < AD_absDev] = np.nan
        gainMap1D_mean = np.nanmean(diff_f1f2_nan, axis=select_correction_axis)

        if select_correction_axis == 1:
            FPN2D = np.repeat(gainMap1D_mean[:, np.newaxis], diff.shape[select_correction_axis],
                              axis=select_correction_axis)
        elif select_correction_axis == 0:
            FPN2D = np.repeat(gainMap1D_mean[np.newaxis, :], diff.shape[select_correction_axis],
                              axis=select_correction_axis)

        return diff - FPN2D


class ColumnProjectionFPNc(QRunnable):

    def __init__(self, video, select_correction_axis, flag_GUI=False):
        """
        This class uses a heuristic procedure called Column Projection FPN (cpFPN) to reduce fixed pattern noise (FPN).

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        select_correction_axis: int (0/1)
            * `0`: FPN will be applied row-wise.
            * `1`: FPN will be applied column-wise.
        """
        super(ColumnProjectionFPNc, self).__init__()
        self.cpu = CPUConfigurations()
        self.video = video
        self.select_correction_axis = select_correction_axis
        self.flag_GUI = flag_GUI
        self.threadpool = QThreadPool()
        self.signals = WorkerSignals()

    @Slot()
    def run(self, *args, **kwargs):
        result = self.cpFPNc(self.select_correction_axis)
        self.signals.result_FPNc.emit(result)

    def cpFPNc(self, select_correction_axis):
        """
        Using the cpFPN approach on video.

        Parameters
        ----------
        select_correction_axis: int (0/1)

           * `0`: FPN will be applied row-wise.
           * `1`: FPN will be applied column-wise.

        Returns
        -------
        output: NDArray
            Video after using the cpFPNc technique.
        """
        if False:
            print('\nFPN correction with parallel loop --->', end=" ")

            result = Parallel(n_jobs=self.cpu.n_jobs, backend='multiprocessing', verbose=self.cpu.verbose)(
                delayed(self.FPNc_kernel)(self.video[f_, :, :], select_correction_axis) for f_ in tqdm(range(0, self.video.shape[0])))
            output = np.asarray(result)

        else:

            print('\ncpFPN correction without parallel loop --->', end=" ")
            output = np.zeros_like(self.video)

            for i_ in tqdm(range(0, self.video.shape[0])):

                if self.flag_GUI is True:
                    self.signals.updateProgress_FPNC.emit(i_)

                diff = self.video[i_, :, :]

                gainMap1D_median = np.median(diff, axis=select_correction_axis)

                if select_correction_axis == 1:
                    FPN2D_median = np.repeat(gainMap1D_median[:, np.newaxis], diff.shape[select_correction_axis],
                                             axis=select_correction_axis)
                elif select_correction_axis == 0:
                    FPN2D_median = np.repeat(gainMap1D_median[np.newaxis, :], diff.shape[select_correction_axis],
                                             axis=select_correction_axis)

                AD_absDev = np.abs(diff - FPN2D_median)
                MAD_1D_MedianAD = np.median(AD_absDev, axis=select_correction_axis) # convert to median from here

                if select_correction_axis == 1:
                    MAD_2D = np.repeat(MAD_1D_MedianAD[:, np.newaxis], diff.shape[select_correction_axis],
                                       axis=select_correction_axis)
                elif select_correction_axis == 0:
                    MAD_2D = np.repeat(MAD_1D_MedianAD[np.newaxis, :], diff.shape[select_correction_axis],
                                       axis=select_correction_axis)

                diff_f1f2_nan = diff.copy()
                diff_f1f2_nan[MAD_2D < AD_absDev] = np.nan
                gainMap1D_mean = np.nanmean(diff_f1f2_nan, axis=select_correction_axis)

                if select_correction_axis == 1:
                    FPN2D = np.repeat(gainMap1D_mean[:, np.newaxis], diff.shape[select_correction_axis],
                                      axis=select_correction_axis)
                elif select_correction_axis == 0:
                    FPN2D = np.repeat(gainMap1D_mean[np.newaxis, :], diff.shape[select_correction_axis],
                                      axis=select_correction_axis)

                output[i_, :, :] = diff - FPN2D

        if self.flag_GUI is True:
            self.signals.finished_FPNc.emit()
        print('Done')

        return output

    def FPNc_kernel(self, diff, select_correction_axis):

        gainMap1D_median = np.median(diff, axis=select_correction_axis)

        if select_correction_axis == 1:
            FPN2D_median = np.repeat(gainMap1D_median[:, np.newaxis], diff.shape[select_correction_axis],
                                     axis=select_correction_axis)
        elif select_correction_axis == 0:
            FPN2D_median = np.repeat(gainMap1D_median[np.newaxis, :], diff.shape[select_correction_axis],
                                     axis=select_correction_axis)

        AD_absDev = np.abs(diff - FPN2D_median)
        MAD_1D_MedianAD = np.median(AD_absDev, axis=select_correction_axis)

        if select_correction_axis == 1:
            MAD_2D = np.repeat(MAD_1D_MedianAD[:, np.newaxis], diff.shape[select_correction_axis],
                               axis=select_correction_axis)
        elif select_correction_axis == 0:
            MAD_2D = np.repeat(MAD_1D_MedianAD[np.newaxis, :], diff.shape[select_correction_axis],
                               axis=select_correction_axis)

        diff_f1f2_nan = diff.copy()
        diff_f1f2_nan[MAD_2D < AD_absDev] = np.nan
        gainMap1D_mean = np.nanmean(diff_f1f2_nan, axis=select_correction_axis)

        if select_correction_axis == 1:
            FPN2D = np.repeat(gainMap1D_mean[:, np.newaxis], diff.shape[select_correction_axis],
                              axis=select_correction_axis)
        elif select_correction_axis == 0:
            FPN2D = np.repeat(gainMap1D_mean[np.newaxis, :], diff.shape[select_correction_axis],
                              axis=select_correction_axis)

        return diff - FPN2D


class FrequencyFPNc:

    def __init__(self, video, inter_flag_parallel_active=True):
        """
        This class corrects FPN using two well-known frequency domain techniques from the literature.

        References
        ----------
        [1] Cao, Yanlong, et al. "A multi-scale non-uniformity correction method based on wavelet decomposition and
        guided filtering for uncooled long wave infrared camera." Signal Processing: Image Communication 60 (2018): 13-21.

        [2] Zeng, Qingjie, et al. "Single infrared image-based stripe non-uniformity correction via a two-stage
        filtering method." Sensors 18.12 (2018): 4299.

        Parameters
        ----------
        video: NDArray
            The video is 3D-numpy (number of frames, width, height).

        inter_flag_parallel_active: bool
            If the user wants to enable generic parallel tasks in CPU configuration, this flag is used to disable parallel execution of this function.
        """
        super(FrequencyFPNc, self).__init__()
        self.cpu = CPUConfigurations()
        self.inter_flag_parallel_active = inter_flag_parallel_active

        self.coeffs = None
        self.arr = None
        self.coeff_slices = None
        self.video_out =None
        self.im_fft_noiseless = None
        self.im_fft_noise = None
        self.im_recon = None
        self.coeffs_result = None
        self.print_color = PrintColors()
        self.video = video

    def update_fFPN(self,  direction='Horizontal', max_iterations=10, width=1):
        """
        This method corrects the FPNc by using FFT [2].

        Parameters
        ----------
        direction: str
            Axis that FPN correction should apply on it.

            * ``'Horizontal'``
            * ``'Vertical'``

        max_iterations: int
            Total number of filtering iterations.

        width: int
            The frequency mask's width.

        Returns
        -------
        n_video: NDArray
            Video after using the fFPNc technique.
        """
        self.max_iterations = max_iterations
        self.video_out = np.ndarray(self.video.shape)
        self.direction = direction

        w = width
        if self.direction == 'Horizontal':
            self.mask = np.ones(self.video[0, :, :].shape)
            center_freq_line_H = round(1 + (0.5 * self.mask.shape[1]))
            self.mask[center_freq_line_H - w:center_freq_line_H + w, :] = 0

        elif self.direction == 'Vertical':
            self.mask = np.ones(self.video[0, :, :].shape)
            center_freq_line_v = round(1 + (0.5 * self.mask.shape[1]))
            self.mask[:, center_freq_line_v - w:center_freq_line_v + w] = 0

        if self.cpu.parallel_active is True and self.inter_flag_parallel_active is True:
            print("\n---start fFPNc with Parallel---")
            result = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(
                delayed(self.fft_kernel)(f_) for f_ in tqdm(range(self.video.shape[0])))

            self.video_out = np.asarray(result)
        else:
            print("\n---start fFPNc without Parallel---")
            result = [self.fft_kernel(f_) for f_ in tqdm(range(0, self.video.shape[0]))]
            self.video_out = np.asarray(result)

        return self.video_out

    def fft_kernel(self, f_):
        selected_frame = self.video[f_, :, :]
        im_fft = np.fft.fftshift(np.fft.fft2(selected_frame))
        amplitude = np.abs(im_fft)
        phase = np.angle(im_fft)

        amplitude_noiseless = np.multiply(self.mask, amplitude)
        im_fft_noiseless = np.multiply(amplitude_noiseless, np.exp(phase * 1j))
        im_fft_noiseless = np.fft.fftshift(im_fft_noiseless)
        im_noiseless = np.fft.ifft2(im_fft_noiseless)

        im_noise = self.video[f_, :, :] - np.real(im_noiseless)

        if self.direction == 'Horizontal':
            selem = rectangle(nrows=1, ncols=5)

        elif self.direction == 'Vertical':
            selem = rectangle(nrows=5, ncols=1)

        filter_img = im_noise.copy()
        for i_ in range(self.max_iterations):
            try:
                filter_img = filters.gaussian(filter_img, sigma=1.2, preserve_range=True)
                # filter_img = rank.mean(filter_img, selem=selem)

            except:
                print(f"{self.WARNING}\nThe Gaussian filter can not work!{self.ENDC}")

        img_FPNc = np.real(im_noiseless) + np.real(filter_img)

        return img_FPNc

    def update_wFPN(self, direction='Horizontal'):
        """
        This method corrects the FPNc by using wavelet[1].

        Parameters
        ----------
        direction: str
            Axis that FPN correction should apply on it.

            * ``'Horizontal'``
            * ``'Vertical'``

        Returns
        -------
        n_video: NDArray
            Video after using the wFPNc technique.
        """
        self.video_out = np.empty_like(self.video)

        if self.cpu.parallel_active is True and self.inter_flag_parallel_active is True:
            print("\n---start wFPNc with Parallel---")
            result = Parallel(n_jobs=self.cpu.n_jobs, backend=self.cpu.backend, verbose=self.cpu.verbose)(
                delayed(self.wavelet_denoise)(f_, direction) for f_ in tqdm(range(self.video.shape[0])))

            self.video_out = np.asarray(result)
        else:
            print("\n---start wFPNc without Parallel---")
            for f_ in tqdm(range(0, self.video.shape[0])):
                self.video_out[f_, :, :] = self.wavelet_denoise(f_, direction=direction)

        return self.video_out

    def wavelet_denoise(self, f_, direction='Horizontal'):

        coeffs_1 = pywt.dwt2(self.video[f_, :, :], wavelet='sym8')
        cA_1, (cH_1, cV_1, cD_1) = coeffs_1

        coeffs_2 = pywt.dwt2(cA_1, wavelet='sym8')
        cA_2, (cH_2, cV_2, cD_2) = coeffs_2

        coeffs_3 = pywt.dwt2(cA_2, wavelet='sym8')
        cA_3, (cH_3, cV_3, cD_3) = coeffs_3

        if direction == 'Horizontal':
            cH_3_corr = self.guided_filter(cH_3, cA_3, win_size=(1, np.int(0.5 * cA_3.shape[0])), eps=0.3**2)
            coeffs_3_corr = cA_3, (cH_3_corr, cV_3, cD_3)
            cA_2_new_ = pywt.idwt2(coeffs_3_corr, 'sym8')
            cA_2_new = self.size_handling(cH_2, cA_2_new_)

            cH_2_corr = self.guided_filter(cH_2, cA_2_new, win_size=(1, np.int(0.25 * cA_3.shape[0])), eps=0.2**2)
            coeffs_2_corr = cA_2_new, (cH_2_corr, cV_2, cD_2)
            cA_1_new_ = pywt.idwt2(coeffs_2_corr, 'sym8')
            cA_1_new = self.size_handling(cH_1, cA_1_new_)

            cH_1_corr = self.guided_filter(cH_1, cA_1_new, win_size=(1, np.int(0.1 * cA_3.shape[0])), eps=0.1**2)
            coeffs_1_corr = cA_1_new, (cH_1_corr, cV_1, cD_1)
            im_corr = pywt.idwt2(coeffs_1_corr, 'sym8')
            return self.size_handling(self.video[f_, :, :], im_corr)

        elif direction == 'Vertical':
            cV_3_corr = self.guided_filter(cV_3, cA_3, win_size=(np.int(0.5 * cA_3.shape[0]), 1), eps=0.3 ** 2)
            coeffs_3_corr = cA_3, (cH_3, cV_3_corr, cD_3)
            cA_2_new_ = pywt.idwt2(coeffs_3_corr, 'sym8')
            cA_2_new = self.size_handling(cV_2, cA_2_new_)

            cV_2_corr = self.guided_filter(cV_2, cA_2_new, win_size=(np.int(0.25 * cA_3.shape[0]), 1), eps=0.2 ** 2)
            coeffs_2_corr = cA_2_new, (cH_2, cV_2_corr, cD_2)
            cA_1_new_ = pywt.idwt2(coeffs_2_corr, 'sym8')
            cA_1_new = self.size_handling(cV_1, cA_1_new_)

            cV_1_corr = self.guided_filter(cV_1, cA_1_new, win_size=(np.int(0.1 * cA_3.shape[0]), 1), eps=0.1 ** 2)
            coeffs_1_corr = cA_1_new, (cH_1, cV_1_corr, cD_1)
            im_corr = pywt.idwt2(coeffs_1_corr, 'sym8')
            return self.size_handling(self.video[f_, :, :], im_corr)

    def guided_filter(self, input_image, guided_image, win_size, eps):
        GF_ = GuidedFilter(guided_image, radius=win_size, eps=eps)
        return GF_.filter(input_image)

    def size_handling(self, im_1, im_2):

        if im_1.shape[0] == im_2.shape[0] and im_1.shape[1] == im_2.shape[1]:
            im_out = im_2
        elif im_1.shape[0] == im_2.shape[0]:
            num_cut = np.abs(im_1.shape[1] - im_2.shape[1])
            im_out = im_2[:, 0:-num_cut]
        elif im_1.shape[1] == im_2.shape[1]:
            num_cut = np.abs(im_1.shape[0] - im_2.shape[0])
            im_out = im_2[0:-num_cut, :]
        else:
            num_cut_0 = np.abs(im_1.shape[0] - im_2.shape[0])
            num_cut_1 = np.abs(im_1.shape[1] - im_2.shape[1])
            im_out = im_2[0:-num_cut_0, 0:-num_cut_1]
        return im_out


from __future__ import print_function
import numpy as np
import math

from math import sqrt
from scipy import spatial
from skimage.feature import peak_local_max
from skimage import img_as_float
import matplotlib.pyplot as plt


def _compute_disk_overlap(d, r1, r2):
    """
    Compute fraction of surface overlap between two disks of radii
    ``r1`` and ``r2``, with centers separated by a distance ``d``.

    Parameters
    ----------
    d : float
      Distance between centers.

    r1 : float
      Radius of the first disk.

    r2 : float
      Radius of the second disk.

    Returns
    -------
    fraction: float
      Fraction of area of the overlap between the two disks.
    """

    ratio1 = (d ** 2 + r1 ** 2 - r2 ** 2) / (2 * d * r1)
    ratio1 = np.clip(ratio1, -1, 1)
    acos1 = math.acos(ratio1)

    ratio2 = (d ** 2 + r2 ** 2 - r1 ** 2) / (2 * d * r2)
    ratio2 = np.clip(ratio2, -1, 1)
    acos2 = math.acos(ratio2)

    a = -d + r2 + r1
    b = d - r2 + r1
    c = d + r2 - r1
    d = d + r2 + r1
    area = (r1 ** 2 * acos1 + r2 ** 2 * acos2 -
          0.5 * sqrt(abs(a * b * c * d)))
    return area / (math.pi * (min(r1, r2) ** 2))


def _compute_sphere_overlap(d, r1, r2):
    """Compute volume overlap fraction between two spheres of radii
    ``r1`` and ``r2``, with centers separated by a distance ``d``.

    Parameters
    ----------
    d : float
        Distance between centers.

    r1 : float
        Radius of the first sphere.

    r2 : float
        Radius of the second sphere.

    Returns
    -------
    fraction: float
        Fraction of volume of the overlap between the two spheres.

    Notes
    -----
    See for example http://mathworld.wolfram.com/Sphere-SphereIntersection.html
    for more details.
    """
    vol = (math.pi / (12 * d) * (r1 + r2 - d) ** 2 *
           (d ** 2 + 2 * d * (r1 + r2) - 3 * (r1 ** 2 + r2 ** 2) + 6 * r1 * r2))
    return vol / (4. / 3 * math.pi * min(r1, r2) ** 3)

def _blob_overlap(blob1, blob2):
    """Finds the overlapping area fraction between two blobs.

    Returns a float representing fraction of overlapped area.

    Parameters
    ----------
    blob1 : sequence of arrays
      A sequence of ``(row, col, sigma)`` or ``(pln, row, col, sigma)``,
      where ``row, col`` (or ``(pln, row, col)``) are coordinates
      of blob and ``sigma`` is the standard deviation of the Gaussian kernel
      which detected the blob.

    blob2 : sequence of arrays
      A sequence of ``(row, col, sigma)`` or ``(pln, row, col, sigma)``,
      where ``row, col`` (or ``(pln, row, col)``) are coordinates
      of blob and ``sigma`` is the standard deviation of the Gaussian kernel
      which detected the blob.

    Returns
    -------
    f : float
      Fraction of overlapped area (or volume in 3D).
    """
    n_dim = len(blob1) - 1
    root_ndim = sqrt(n_dim)

    r1 = blob1[-1] * root_ndim
    r2 = blob2[-1] * root_ndim

    d = sqrt(np.sum((blob1[:-1] - blob2[:-1]) ** 2))
    if d > r1 + r2:
        return 0

    if d <= abs(r1 - r2):
        return 1

    if n_dim == 2:
        return _compute_disk_overlap(d, r1, r2)

    else:
        return _compute_sphere_overlap(d, r1, r2)


def _prune_blobs(blobs_array, overlap):
    """Eliminated blobs with area overlap.

    Parameters
    ----------
    blobs_array : NDArray
      A 2d array with each row representing 3 (or 4) values,
      ``(row, col, sigma)`` or ``(pln, row, col, sigma)`` in 3D,
      where ``(row, col)`` (``(pln, row, col)``) are coordinates of the blob
      and ``sigma`` is the standard deviation of the Gaussian kernel which
      detected the blob.
      This array must not have a dimension of size 0.

    overlap : float
      A value between 0 and 1. If the fraction of area overlapping for 2
      blobs is greater than `overlap` the smaller blob is eliminated.

    Returns
    -------
    A :(NDArray
      `array` with overlapping blobs removed.
    """
    sigma = blobs_array[:, -1].max()
    distance = 2 * sigma * sqrt(blobs_array.shape[1] - 1)
    tree = spatial.cKDTree(blobs_array[:, :-1])
    pairs = np.array(list(tree.query_pairs(distance)))
    if len(pairs) == 0:
        return blobs_array
    else:
        for (i, j) in pairs:
          blob1, blob2 = blobs_array[i], blobs_array[j]
          if _blob_overlap(blob1, blob2) > overlap:
            if blob1[-1] > blob2[-1]:
              blob2[-1] = 0
            else:
              blob1[-1] = 0

    return np.array([b for b in blobs_array if b[-1] > 0])


def blob_frst(image, min_radial=1, max_radial=50, radial_step=1.6, threshold=2.0, alpha=2, beta=1e-3, stdFactor=4,
              mode='BOTH', overlap=.5, *, exclude_border=False):
    """
    This function uses scikit-image's local maximum function to localize the PSFs using a ``_frst``.

    Parameters
    ----------
    img: NDArray
       Input_video image, grayscale.

    alpha: float
       Strictness of symmetry transform (higher=more strict; 2 is good place to start)

    beta: float
        Gradient threshold_min parameter, float in [0,1]

    stdFactor: float
       Standard deviation factor for gaussian kernel

    mode: str
       BRIGHT, DARK, or BOTH

    min_radial: int
       integer value for radius size in pixels (n in the original paper); also is used as gaussian kernel size

    max_radial: int
       integer value for radius size in pixels (n in the original paper); also is used as gaussian kernel size

    threshold: float
        The absolute lower bound for scale space maxima. Local maxima smaller than thresh are ignored. Reduce this
        to detect blobs with less intensities.

    overlap: float
        A value between 0 and 1. If the area of two blobs are overlapping by a fraction greater than threshold_min, smaller blobs are eliminated.

    exclude_border: int, tuple of ints, or bool, optional
        If positive integer, `exclude_border` excludes peaks from within
        `exclude_border`-pixels of the border of the image.
        If tuple of non-negative ints, the length of the tuple must match the
        input array's dimensionality.  Each element of the tuple will exclude
        peaks from within `exclude_border`-pixels of the border of the image
        along that dimension.
        If True, takes the `min_distance` parameter as value.
        If zero or False, peaks are identified regardless of their distance
        from the border.

    Returns
    -------
    local_maxima_: list
        [y, x, sigma]

    """

    image = img_as_float(image)

    # Gaussian filter requires that sequence-bin_type sigmas have same
    # dimensionality as image. This broadcasts scalar kernels
    if np.isscalar(max_radial):
      max_radial = np.full(image.ndim, max_radial, dtype=float)
    if np.isscalar(min_radial):
      min_radial = np.full(image.ndim, min_radial, dtype=float)

    # Convert sequence types to array
    min_radial = np.asarray(min_radial, dtype=float)
    max_radial = np.asarray(max_radial, dtype=float)

    # a geometric progression of standard deviations for gaussian kernels
    radial_list = np.array([radial for radial in range(int(min_radial[0]), int(max_radial[0] + 1), int(radial_step))])

    frst_images = [frst(image, radii=r, alpha=alpha, beta=beta, stdFactor=stdFactor, mode=mode) for r in radial_list]

    image_cube = np.stack(frst_images, axis=-1)
    image_cube = np.power(image_cube, 2)
    # image_cube = np.std(image_cube, axis=2)

    # local_maxima = get_local_maxima(image_cube, threshold_min)
    local_maxima = peak_local_max(image_cube, threshold_abs=threshold,
                                  footprint=np.ones((3,) * (image.ndim + 1)),
                                  threshold_rel=0.0,
                                  exclude_border=exclude_border)
    # Catch no peaks
    if local_maxima.size == 0:
      return np.empty((0, 3))

    local_maxima_ = []

    for idx, l_m in enumerate(local_maxima):
        local_maxima_.append([l_m[0], l_m[1], radial_list[l_m[2]]])

    return local_maxima_






from git import Repo
from pathlib import Path
from os.path import exists
import os
import sys
import subprocess

from piscat.Visualization.print_colors import PrintColors


def parse_args():
    """Parses the command line arguments.

    The first command line argument is the module name.

    Raises
    ------
    Exception
        if command line arguments are invalid (e.g. more than one
        argument)

    """

    if len(sys.argv) == 2:
        name = sys.argv[1]
    else:
        raise Exception('ERROR! Please specify plugin name')
    return name


class Plugin(PrintColors):
    def __init__(self):
        self.dic_git_url = {'UAI': 'https://github.com/SandoghdarLab/PiSCAT_plugin_UAI.git'}

    def confirm(self, prompt):
        """Get user confirmation from standard input.

        Parameters
        ----------
        prompt: str
            The message to show.

        Returns
        -------
        True or false depending on whether the user has answered with yes or no.
        """
        answer = ''
        while answer.lower() != 'y' and answer.lower() != 'n':
            answer = input(prompt)
        return answer == 'y'

    def copy_plugin(self, plu_name):
        """Copies the PiSCAT plugin to a plugin directory.

        Parameters
        ----------
        plu_name: name of the plugin

        Returns
        -------
        target: boolean
            Specify and plugin data changed or not
        """
        target = os.path.dirname(os.path.realpath(__file__))
        target = os.path.join(target, plu_name)
        if exists(target):
            print('Warning! Target directory %s already exists.' % str(target))
            answer = self.confirm(f'{self.BOLD}\n Target directory %s already exists. '
                                  f'Do you want to pull the latest change (y/n): {self.ENDC}' % str(target))
            if answer:
                try:
                    repo = Repo(target)
                    o = repo.remotes.origin
                    o.pull()
                    print("The plugin repository pulled successfully")
                    return True
                except:
                    print("The plugin repository can not be pulled")
                    return False
            else:
                return False

        else:
            try:
                Repo.clone_from(self.dic_git_url[plu_name], target)
                print('Target directory %s was successfully cloned.' % str(target))
                return True
            except:
                print("The plugin repository can not be cloned")
                return False

    def add_plugin(self, plugin_name):

        if plugin_name == 'UAI':
            answer = self.confirm(f'{self.BOLD}\n This plugin is not obeyed the general PiSCAT license, and it has its specific {self.ENDC}'
                                  f'{self.BOLD}license. please visit "https://github.com/SandoghdarLab/PiSCAT_plugin" for more {self.ENDC}'
                                  f'{self.BOLD}information.{self.ENDC} {self.OKGREEN}\nI agree to install this plugin with a new License (y/n):{self.ENDC}')
            if answer:
                files_changes = self.copy_plugin(plugin_name)
                if files_changes:
                    folder_path = os.path.dirname(os.path.realpath(__file__))
                    path = os.path.join(folder_path, plugin_name)
                    # implement pip as a subprocess:
                    subprocess.run('pip install -r' + path + '\\requirements.txt')

    # def start_tutorials():
    #     """Copy the PiSCAT tutorials to the target directory and start Jupyter server
    #
    #     The target directory is the first argument of sys.argv.
    #     """
    #
    #     target_root = parse_args()
    #     root = copy_tutorials(target_root)
    #     jupyter = find_jupyter()
    #     start_server(root, jupyter)


def main(p_name):
    p_ = Plugin()
    p_.add_plugin(p_name)


if __name__ == '__main__':
    plugin_name = parse_args()
    main(plugin_name)



import numpy as np
import scipy.optimize
from ipywidgets import widgets
from ipywidgets import Layout, interact
from matplotlib import pyplot as plt, cm as cm
from matplotlib.patches import Circle, Arrow, Rectangle
from mpl_toolkits.axes_grid1 import make_axes_locatable
from scipy.ndimage import median_filter
from skimage.filters import median
from skimage.morphology import disk

from piscat.Localization import gaussian_2D_fit
from piscat.Localization import directional_intensity
from piscat.Trajectory.data_handeling import protein_trajectories_list2dic
from piscat.InputOutput import read_status_line


class JupyterDisplay:

    def __init__(self, video, median_filter_flag=False, color='gray', title=None, xlabel=None, ylabel=None,
                 imgSizex=5, imgSizey=5, extent=None, IntSlider_width='500px', step=1):
        """
        This class displays the video in jupyter notebook.

        Parameters
        ----------
        video:(NDArray)
            Input video.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        title: list
            A list of string titles with a length equal to the number of frames in the video.

        xlabel: str
            The label text for x-axis.

        ylabel: str
            The label text for y-axis.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        extent:
            The extent kwarg determines the bounding box in data coordinates that the image will fill, which is
            specified in data coordinates as (left, right, bottom, top).

        IntSlider_width: str
            Size of slider

        step: int
            Stride between visualization frames.

        """
        self.color = color
        self.video = video
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.title = title
        self.extent = extent

        self.median_filter_flag = median_filter_flag

        self.xlabel = xlabel
        self.ylabel = ylabel

        interact(self.display, frame=widgets.IntSlider(min=0, max=self.video.shape[0] - 1, step=step, value=10,
                                                       layout=Layout(width=IntSlider_width),
                                                       readout_format='10', continuous_update=False,
                                                       description='Frame:'))

    def display(self, frame):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])

        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(frame), :, :], 3)
        else:
            frame_v = self.video[int(frame), :, :]

        if self.extent is not None:
            myplot = ax.imshow(frame_v, cmap=self.color, extent=self.extent)
        else:
            myplot = ax.imshow(frame_v, cmap=self.color)

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(myplot, cax=cax)

        if self.xlabel is not None:
            ax.set_xlabel(self.xlabel)

        if self.ylabel is not None:
            ax.set_ylabel(self.ylabel)

        if self.title is not None:
            ax.set_title(self.title[int(frame)])

        plt.show()


class JupyterDisplay_StatusLine:

    def __init__(self, video, median_filter_flag=False, color='gray', imgSizex=5, imgSizey=5, IntSlider_width='500px', step=1, value=0):
        """
        This class displays the video in the Jupyter notebook interactively while highlight status line

        Parameters
        ----------
        video: NDArray
            Input video.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        IntSlider_width: str
            Size of slider.

        step: int
            Stride between visualization frames.

        value: int
            Initial frame value for visualization
        """
        self.color = color
        self.video = video
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.median_filter_flag = median_filter_flag

        status_ = read_status_line.StatusLine(video)
        self.video_remove, status_information = status_.find_status_line()
        self.statusLine_position = status_information['status_line_position']

        interact(self.display, frame=widgets.IntSlider(min=0, max=self.video.shape[0] - 1, step=step, value=value,
                                                       layout=Layout(width=IntSlider_width),
                                                       readout_format='10', continuous_update=False,
                                                       description='Frame:'))

    def display(self, frame):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])

        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(frame), :, :], 3)
            frame_v_rm = median_filter(self.video_remove[int(frame), :, :], 3)

        else:
            frame_v = self.video[int(frame), :, :]
            frame_v_rm = self.video_remove[int(frame), :, :]


        myplot_rm = ax.imshow(frame_v_rm, cmap=self.color)
        myplot = ax.imshow(frame_v, cmap=self.color)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(myplot_rm, cax=cax)

        if self.statusLine_position == 'column':
            rect = Rectangle((0, self.video.shape[2]), self.video.shape[2], 1, linewidth=10, edgecolor='r',
                                  facecolor='none')
            ax.add_patch(rect)
        elif self.statusLine_position == 'row':
            rect = Rectangle((self.video.shape[1], 0), 1, self.video.shape[1], linewidth=10, edgecolor='r',
                                  facecolor='none')
            ax.add_patch(rect)
        plt.show()


class JupyterPSFs_localizationDisplay:

    def __init__(self, video, df_PSFs, median_filter_flag=False, color='gray', imgSizex=5, imgSizey=5, IntSlider_width='500px', step=1,  value=0):
        """
        This class displays the video in the Jupyter notebook interactively while highlight PSFs.

        Parameters
        ----------
        video: NDArray
           Input video.

        df_PSFs: panda data_frame
            Data Frames that contains the location of PSFs.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        IntSlider_width: str
           Size of slider.

        step: int
           Stride between visualization frames.

        value: int
           Initial frame value for visualization
       """
        self.video = video
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.df_PSFs = df_PSFs
        self.median_filter_flag = median_filter_flag
        self.color = color

        interact(self.show_psf, frame_number=widgets.IntSlider(min=0, max=self.video.shape[0] - 1, step=step, value=value,
                                                               readout_format='1', continuous_update=False, layout=Layout(width=IntSlider_width),
                                                               description='Frame:'))

    def show_psf(self, frame_number):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])
        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(frame_number), :, :], 3)
        else:
            frame_v = self.video[int(frame_number), :, :]

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        myplot = ax.imshow(frame_v, cmap=self.color)
        plt.colorbar(myplot, cax=cax)

        particle = self.df_PSFs.loc[self.df_PSFs['frame'] == frame_number]
        particle_X = particle['x'].tolist()
        particle_Y = particle['y'].tolist()
        particle_sigma = particle['sigma'].tolist()

        for j_ in range(len(particle_X)):
            y = int(particle_Y[j_])
            x = int(particle_X[j_])
            sigma = particle_sigma[j_]
            ax.add_patch(Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor='r', facecolor='none', linewidth=2))
        plt.show()


class JupyterPSFs_localizationPreviewDisplay:

    def __init__(self, video, df_PSFs, title='', frame_num=None, median_filter_flag=False, color='gray', imgSizex=5, imgSizey=5, IntSlider_width='500px', step=1):
        """
        This class displays the video in the Jupyter notebook interactively while highlight PSFs.

        Parameters
        ----------
        video: NDArray
           Input video.

        df_PSFs: panda data_frame
            Data Frames that contains the location of PSFs.

        title: str
            It defines title of plot.

        frame_num: list
            list of frame that we want to see preview of localization.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        IntSlider_width: str
           Size of slider.

        step: int
           Stride between visualization frames.
       """
        self.video = video
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.df_PSFs = df_PSFs
        self.median_filter_flag = median_filter_flag
        self.color = color
        if self.df_PSFs is not None:
            self.frame_num = np.sort(np.unique(df_PSFs['frame'].tolist()))
        else:
            self.frame_num = frame_num
        self.currentFrame = 0
        self.IntSlider_width = IntSlider_width
        self.step = step
        self.title = title

    def display_run(self):
        interact(self.show_psf,
                 index=widgets.IntSlider(min=0, max=len(self.frame_num) - 1, step=self.step, value=self.currentFrame,
                                         readout_format='1', continuous_update=False,
                                         layout=Layout(width=self.IntSlider_width),
                                         description='index:'))

    def show_psf(self, index):
        frame_number = self.frame_num[index]
        self.currentFrame = frame_number
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])
        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(frame_number), :, :], 3)
        else:
            frame_v = self.video[int(frame_number), :, :]

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        myplot = ax.imshow(frame_v, cmap=self.color)
        ax.set_title(self.title)
        plt.colorbar(myplot, cax=cax)

        if self.df_PSFs is not None:
            particle = self.df_PSFs.loc[self.df_PSFs['frame'] == frame_number]
            particle_X = particle['x'].tolist()
            particle_Y = particle['y'].tolist()
            particle_sigma = particle['sigma'].tolist()

            for j_ in range(len(particle_X)):
                y = int(particle_Y[j_])
                x = int(particle_X[j_])
                sigma = particle_sigma[j_]
                ax.add_patch(Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor='r', facecolor='none', linewidth=2))
        plt.show()

    def set_df(self, new_df):
        self.df_PSFs = new_df


class JupyterPSFs_subplotLocalizationDisplay:

    def __init__(self, list_videos, list_df_PSFs, numRows, numColumns, list_titles=None,
                 median_filter_flag=False, color='gray', imgSizex=5, imgSizey=5, IntSlider_width='500px', step=1, value=0):
        """
        This class shows several videos (with the same number of frames) at once in the Jupyter notebook interactively while highlight localize PSFs.

        Parameters
        ----------
        list_videos: list of NDArray
          List of videos.

        list_df_PSFs: list panda data_frame
           List Data Frames that contains the location of PSFs for each video.

        numRows: int
            It defines number of rows in sub-display.

        numColumns: int
            It defines number of columns in sub-display.

        list_titles: list str
            List of titles for each sub plot.

        median_filter_flag: bool
          In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
          It defines the colormap for visualization.

        imgSizex: int
          Image length size.

        imgSizey: int
          Image width size.

        IntSlider_width: str
          Size of slider

        step: int
          Stride between visualization frames.

        value: int
            Initial frame value for visualization
        """
        self.list_video = list_videos
        self.numVideos = len(list_videos)
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.list_df_PSFs = list_df_PSFs
        self.median_filter_flag = median_filter_flag
        self.color = color
        self.numRows = numRows
        self.numColumns = numColumns

        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        max_numberFrames = np.max([vid_.shape[0] for vid_ in list_videos])
        interact(self.show_psf, frame_number=widgets.IntSlider(min=0, max=max_numberFrames, step=step, value=value,
                                                               readout_format='1', continuous_update=False, layout=Layout(width=IntSlider_width),
                                                               description='Frame:'))

    def show_psf(self, frame_number):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        grid = plt.GridSpec(self.numRows, self.numColumns, hspace=0.3, wspace=0.3)

        imgGrid_list = []
        for i in range(self.numRows):
            for j in range(self.numColumns):
                imgGrid_list.append(fig.add_subplot(grid[i, j]))

        for img_, tit_, img_grid_, df_PSFs in zip(self.list_video, self.list_titles, imgGrid_list, self.list_df_PSFs):

            if self.median_filter_flag:

                frame_v = median_filter(img_[int(frame_number), :, :], 3)
            else:
                frame_v = img_[int(frame_number), :, :]

            divider = make_axes_locatable(img_grid_)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            myplot = img_grid_.imshow(frame_v, cmap=self.color)
            img_grid_.axis('off')
            plt.colorbar(myplot, cax=cax)
            if tit_ is not None:
                img_grid_.set_title(tit_)

            particle = df_PSFs.loc[df_PSFs['frame'] == frame_number]
            particle_X = particle['x'].tolist()
            particle_Y = particle['y'].tolist()
            particle_sigma = particle['sigma'].tolist()

            for j_ in range(len(particle_X)):
                y = int(particle_Y[j_])
                x = int(particle_X[j_])
                sigma = particle_sigma[j_]
                img_grid_.add_patch(Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor='r', facecolor='none', linewidth=2))
        plt.show()


class JupyterPSFs_TrackingDisplay:

    def __init__(self, video, df_PSFs, median_filter_flag=False, step=1, color='gray', imgSizex=5, imgSizey=5, value=0):
        """
        This class displays video in the Jupyter notebook interactively while highlighting PSFs with trajectories.

        Parameters
        ----------
        video: NDArray
           Input video.

        df_PSFs: panda data_frame
            Data Frames that contains the location of PSFs.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        IntSlider_width: str
           Size of slider.

        step: int
           Stride between visualization frames.

        value: int
            Initial frame value for visualization
        """

        self.video = video
        self.df_PSFs = df_PSFs
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.df_PSFs = df_PSFs
        self.median_filter_flag = median_filter_flag
        self.color = color
        self.median_filter_flag = median_filter_flag
        if self.df_PSFs.particle.isnull().any():
            self.df_PSFs['particle'] = 0
            self.list_particles_idx = self.df_PSFs.particle.unique()
        else:
            self.list_particles_idx = self.df_PSFs.particle.unique()

        colors_ = cm.autumn(np.linspace(0, 1, len(self.list_particles_idx)))

        self.colors = colors_[0:len(self.list_particles_idx), :]
        interact(self.show_psf, frame_number=widgets.IntSlider(min=0, max=self.video.shape[0] - 1, step=step, value=value,
                    readout_format='1', continuous_update=False, description='Frame:'))

    def show_psf(self, frame_number):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])

        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(frame_number), :, :], 3)
        else:
            frame_v = self.video[int(frame_number), :, :]

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        myplot = ax.imshow(frame_v, cmap=self.color)
        plt.colorbar(myplot, cax=cax)

        particle = self.df_PSFs.loc[self.df_PSFs['frame'] == frame_number]
        particle_X = particle['x'].tolist()
        particle_Y = particle['y'].tolist()
        particle_sigma = particle['sigma'].tolist()
        particle_labels = particle['particle'].tolist()

        if len(particle_labels) != 0:

            for j_, p_l in zip(range(len(particle_X)), particle_labels):
                y = int(particle_Y[j_])
                x = int(particle_X[j_])
                sigma = particle_sigma[j_]
                ax = plt.gca()
                ax.add_patch(Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor=self.colors[int(p_l)], facecolor='none', linewidth=2))

                all_particle_ = self.df_PSFs.loc[self.df_PSFs['particle'] == p_l]
                particle_f_ = all_particle_['frame'].tolist()
                particle_X_= all_particle_['x'].tolist()
                particle_Y_ = all_particle_['y'].tolist()
                particle_sigma_ = all_particle_['sigma'].tolist()
                for f_, x_, y_, sigma_ in zip(particle_f_, particle_X_, particle_Y_, particle_sigma_):
                    if f_ <= frame_number:
                        ax.add_patch(Circle((int(x_), int(y_)), radius=0.1, edgecolor=self.colors[int(p_l)], facecolor='none', linewidth=1))
                        # ax.add_collection((x_, y_), autolim=True, cmap=self.colors [p_l])
        plt.show()


class JupyterSelectedPSFs_localizationDisplay:

    def __init__(self, video, particles, particles_num='#0', frame_extend=0, median_filter_flag=False, flag_fit2D=False,
                 color='gray', imgSizex=10, imgSizey=10, IntSlider_width='500px', step=1, value=0):
        """
        This class interactively shows video in a Jupyter notebook while highlighting PSFs based on ID.

        Parameters
        ----------
        video: NDArray
           Input video.

        particles: dic
             Dictionary similar to the following structures.:

            | {"#0": {'intensity_horizontal': ..., 'intensity_vertical': ..., ..., 'particle_ID': ...},
                "#1": {}, ...}

        particles_num: str
            Choose the corresponding key in the particles dictionary.

        frame_extend: int
            Display particle for ``frame_extend`` before and after segmented ones. In case there are not enough frames before/after
            , it shows only for the number of existing frames.

        flag_fit2D: bool
            It activate 2D-Gaussian fit to extract fitting information of selected PSF.

        median_filter_flag: bool
           In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
           It defines the colormap for visualization.

        imgSizex: int
           Image length size.

        imgSizey: int
           Image width size.

        IntSlider_width: str
           Size of slider.

        step: int
           Stride between visualization frames.

        value: int
            Initial frame value for visualization
        """

        if type(particles) is list:
            particles = protein_trajectories_list2dic(particles)

        if type(particles) is dict:
            for key, item in particles.items():
                if key == particles_num:
                    intensity_horizontal = item['intensity_horizontal']
                    intensity_vertical = item['intensity_vertical']
                    center_int = item['center_int']
                    center_int_flow = item['center_int_flow']
                    self.frame_number_ = item['frame_number']
                    self.sigma = item['sigma']
                    self.x_center = item['x_center']
                    self.y_center = item['y_center']
                    self.particle_ID = item['particle_ID']

        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.video = video
        self.color = color
        self.flag_fit2D = flag_fit2D
        self.median_filter_flag = median_filter_flag

        video_frameNum = video.shape[0]
        max_particle_frame = np.max(self.frame_number_)
        min_particle_frame = np.min(self.frame_number_)

        max_extend_particle_frame = np.min([max_particle_frame+frame_extend, video_frameNum])
        min_extend_particle_frame = np.max([min_particle_frame-frame_extend, 0])

        self.frame_number = list(range(min_extend_particle_frame, max_extend_particle_frame, 1))
        interact(self.show_psf,
                 f_num=widgets.IntSlider(min=self.frame_number[0], max=self.frame_number[-1] - 1, step=step, value=value,
                                         readout_format='1', continuous_update=False, layout=Layout(width=IntSlider_width),
                                         description='Frame:'))

    def show_psf(self, f_num):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        ax = fig.add_axes([0, 0, 1, 1])
        if self.flag_fit2D:
            fit_params = self.fit_2D_gussian(frame_num=f_num, scale=5)

        if self.median_filter_flag:
            frame_v = median_filter(self.video[int(f_num), :, :], 3)
        else:
            frame_v = self.video[int(f_num), :, :]

        myplot = ax.imshow(frame_v, cmap=self.color)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(myplot, cax=cax)

        if f_num <= np.max(self.frame_number_) and f_num >= np.min(self.frame_number_):
            idx_ = np.where(self.frame_number_ == f_num)

            particle_X = self.x_center[idx_]
            particle_Y = self.y_center[idx_]
            particle_sigma = self.sigma[idx_]
            particle_labels = self.particle_ID[idx_]

            y = int(particle_Y)
            x = int(particle_X)
            sigma = particle_sigma
            ax.add_patch(Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor='r', facecolor='none', linewidth=2))

        plt.show()

    def fit_2D_gussian(self, frame_num, scale=5):

        particle = self.list_psf.loc[self.list_psf['frame'] == frame_num]
        particle_X = particle['x'].tolist()
        particle_Y = particle['y'].tolist()
        particle_sigma = particle['sigma'].tolist()
        particle_center_intensity = particle['center_intensity'].tolist()

        index_list = [index for index in particle.index]

        list_one_frame_fit = []
        if particle.shape[0] > 0:
            for p_x, p_y, sigma_0, c_0, i_ in zip(particle_X, particle_Y, particle_sigma, particle_center_intensity,
                                                  index_list):
                window_size = scale * np.sqrt(2) * sigma_0
                start_sigma = sigma_0
                if p_x > window_size and p_y > window_size:
                    window_frame = self.video[int(frame_num), int(p_y - window_size) + 1:int(p_y + window_size),
                                   int(p_x - window_size) + 1:int(p_x + window_size)]
                    w_s = window_size
                else:
                    window_size_temp = window_size
                    while p_x < window_size_temp or p_y < window_size_temp:
                        window_size_temp = window_size_temp - 2
                    window_frame = self.video[int(frame_num),
                                   int(p_y - window_size_temp) + 1:int(p_y + window_size_temp),
                                   int(p_x - window_size_temp) + 1:int(p_x + window_size_temp)]
                    w_s = window_size_temp

                fit_params_ = gaussian_2D_fit.fit_2D_Gaussian_varAmp(window_frame, sigma_x=start_sigma,
                                                                     sigma_y=start_sigma,
                                                                     display_flag=False)
                fit_params = fit_params_[1]
                fit_errors = fit_params_[2]

                print('Fit Amplitude:', fit_params[0], '\u00b1', fit_errors[0])
                print('Fit X-Center: ', fit_params[1], '\u00b1', fit_errors[1])
                print('Fit Y-Center: ', fit_params[2], '\u00b1', fit_errors[2])
                print('Fit X-Sigma:  ', fit_params[3], '\u00b1', fit_errors[3])
                print('Fit Y-Sigma:  ', fit_params[4], '\u00b1', fit_errors[4])
                print('Fit Bias:  ', fit_params[5], '\u00b1', fit_errors[5])

                fit_params = [frame_num, i_, fit_params_, p_x, p_y, w_s]
                return fit_params


class JupyterSubplotDisplay:

    def __init__(self, list_videos, numRows, numColumns, list_titles=None, imgSizex=20, imgSizey=20, IntSlider_width='500px',
                 median_filter_flag=False, color='gray', step=1, value=0, vmin=None, vmax=None):

        """
        This class interactively displays several videos (with the same number of frames) in a Jupyter notebook.

        Parameters
        ----------
        list_videos: list of NDArray
          List of videos

        numRows: int
            It defines number of rows in sub-display

        numColumns: int
            It defines number of columns in sub-display

        list_titles: list str
            List of titles for each sub plot

        median_filter_flag: bool
          In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        IntSlider_width: str
            Size of slider

        step: int
            Stride between visualization frames.

        value: int
            Initial frame value for visualization
        """

        self.color = color
        self.vmin = vmin
        self.vmax = vmax

        self.video = list_videos
        self.median_filter_flag = median_filter_flag
        self.numRows = numRows
        self.numColumns = numColumns
        self.numVideos = len(list_videos)
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey

        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        max_numberFrames = np.max([vid_.shape[0] for vid_ in list_videos])

        interact(self.display, frame=widgets.IntSlider(min=0, max=max_numberFrames-1, step=step, value=value, layout=Layout(width=IntSlider_width),
                                                       readout_format='100', continuous_update=False, description='Frame:'))

    def display(self, frame):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))

        grid = plt.GridSpec(self.numRows, self.numColumns, hspace=0.3, wspace=0.3)

        imgGrid_list = []
        for i in range(self.numRows):
            for j in range(self.numColumns):
                imgGrid_list.append(fig.add_subplot(grid[i, j]))

        for img_, tit_, img_grid_ in zip(self.video, self.list_titles, imgGrid_list):

            if self.median_filter_flag:
                frame_v = median_filter(img_[int(frame), :, :], 3)
            else:
                frame_v = img_[int(frame), :, :]

            # img_grid_.imshow(frame_v, cmap=self.color)
            # img_grid_.axis('off')

            divider = make_axes_locatable(img_grid_)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            if self.vmin is None or self.vmax is None:
                myplot = img_grid_.imshow(frame_v, cmap=self.color)
            else:
                myplot = img_grid_.imshow(frame_v, cmap=self.color, vmin=self.vmin, vmax=self.vmax)
            img_grid_.axis('off')
            plt.colorbar(myplot, cax=cax)
            if tit_ is not None:
                img_grid_.set_title(tit_)

        plt.show()





class JupyterPSFs_2_modality_subplotLocalizationDisplay:

    def __init__(self, list_videos, list_df_PSFs_1, list_df_PSFs_2, numRows, numColumns, list_titles=None,
                 median_filter_flag=False, color='gray', imgSizex=5, imgSizey=5, IntSlider_width='500px', step=1, value=0,
                 edgecolor_1='r', edgecolor_2='g'):
        """
        This class will interactively sub-display multiple videos in a Jupyter notebook while highlighting PSFs determined in two modalities.

        Parameters
        ----------
        list_videos: list of NDArray
          List of videos.

        list_df_PSFs_1: list
            list of all data Frames that contains the location of PSFs in first modality.

        list_df_PSFs_2: list
            list of all data Frames that contains the location of PSFs in second modality.

        numRows: int
            It defines number of rows in sub-display.

        numColumns: int
            It defines number of columns in sub-display.

        list_titles: list str
            List of titles for each sub plot.

        median_filter_flag: bool
          In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
          It defines the colormap for visualization.

        imgSizex: int
          Image length size.

        imgSizey: int
          Image width size.

        IntSlider_width: str
          Size of slider

        step: int
          Stride between visualization frames.

        value: int
            Initial frame value for visualization

        edgecolor_1: str
            The color of the circle that annotates the first modality on the display.

        edgecolor_2: str
            The color of the circle that annotates the second modality on the display.

        """
        self.list_video = list_videos
        self.numVideos = len(list_videos)
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.list_df_PSFs_1 = list_df_PSFs_1
        self.list_df_PSFs_2 = list_df_PSFs_2
        self.edgecolor_1 = edgecolor_1
        self.edgecolor_2 = edgecolor_2
        self.median_filter_flag = median_filter_flag
        self.color = color
        self.numRows = numRows
        self.numColumns = numColumns

        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        max_numberFrames = np.max([vid_.shape[0] for vid_ in list_videos])
        interact(self.show_psf, frame_number=widgets.IntSlider(min=0, max=max_numberFrames-1, step=step, value=value,
                                                               readout_format='1', continuous_update=False, layout=Layout(width=IntSlider_width),
                                                               description='Frame:'))

    def show_psf(self, frame_number):
        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))
        grid = plt.GridSpec(self.numRows, self.numColumns, hspace=0.3, wspace=0.3)

        imgGrid_list = []
        for i in range(self.numRows):
            for j in range(self.numColumns):
                imgGrid_list.append(fig.add_subplot(grid[i, j]))

        for img_, tit_, img_grid_, df_PSFs_1,  df_PSFs_2 in zip(self.list_video, self.list_titles, imgGrid_list, self.list_df_PSFs_1, self.list_df_PSFs_2):

            if self.median_filter_flag:

                frame_v = median_filter(img_[int(frame_number), :, :], 3)
            else:
                frame_v = img_[int(frame_number), :, :]

            divider = make_axes_locatable(img_grid_)
            cax = divider.append_axes("right", size="5%", pad=0.05)
            myplot = img_grid_.imshow(frame_v, cmap=self.color)
            img_grid_.axis('off')
            plt.colorbar(myplot, cax=cax)
            if tit_ is not None:
                img_grid_.set_title(tit_)

            particle_1 = df_PSFs_1.loc[df_PSFs_1['frame'] == frame_number]
            particle_X_1 = particle_1['x'].tolist()
            particle_Y_1 = particle_1['y'].tolist()
            particle_sigma_1 = particle_1['sigma'].tolist()

            particle_2 = df_PSFs_2.loc[df_PSFs_2['frame'] == frame_number]
            particle_X_2 = particle_2['x'].tolist()
            particle_Y_2 = particle_2['y'].tolist()
            particle_sigma_2 = particle_2['sigma'].tolist()

            for j_ in range(len(particle_X_1)):
                y = int(particle_Y_1[j_])
                x = int(particle_X_1[j_])
                sigma = particle_sigma_1[j_]
                img_grid_.add_patch(Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor=self.edgecolor_1, facecolor='none',
                                    linewidth=2))

            for j_ in range(len(particle_X_2)):
                y = int(particle_Y_2[j_])
                x = int(particle_X_2[j_])
                sigma = particle_sigma_2[j_]
                img_grid_.add_patch(Circle((x, y), radius=np.sqrt(2) * sigma, edgecolor=self.edgecolor_2, facecolor='none',
                                    linewidth=2))

        plt.show()


class JupyterFPNcDisplay:

    def __init__(self, list_videos, list_titles=None, correction_axis=0, numRows=1, numColumns=2,
                 imgSizex=20, imgSizey=20, IntSlider_width='500px',
                 median_filter_flag=False, color='gray', step=1, value=0):
        """
        This class can sub-display several FPNc video in the Jupyter notebook interactively while 1D projection on
        the direction of correction axis illustrates below on each subplot.

        Parameters
        ----------
        list_videos: list of NDArray
          List of videos

        correction_axis: (0/1)
            selecting the axis that FPNc correction should visualize for it.

        numRows: int
            It defines number of rows in sub-display.

        numColumns: int
            It defines number of columns in sub-display.

        list_titles: list str
            List of titles for each sub plot.

        median_filter_flag: bool
            In case it defines as True, a median filter is applied with size 3 to remove hot pixel effect.

        color: str
            It defines the colormap for visualization.

        imgSizex: int
            Image length size.

        imgSizey: int
            Image width size.

        IntSlider_width: str
            Size of slider

        step: int
            Stride between visualization frames.

        value: int
            Initial frame value for visualization
        """
        self.color = color
        self.video = list_videos
        self.median_filter_flag = median_filter_flag
        self.numRows = numRows
        self.numColumns = numColumns
        self.numVideos = len(list_videos)
        self.imgSizex = imgSizex
        self.imgSizey = imgSizey
        self.correction_axis = correction_axis

        if list_titles is None:
            self.list_titles = [None for _ in range(self.numVideos)]
        else:
            self.list_titles = list_titles

        max_numberFrames = np.max([vid_.shape[0] for vid_ in list_videos])

        interact(self.display, frame=widgets.IntSlider(min=0, max=max_numberFrames, step=step, value=value, layout=Layout(width=IntSlider_width),
                                                       readout_format='100', continuous_update=False,
                                                       description='Frame:'))

    def display(self, frame):
        self.fig = plt.figure(figsize=(self.imgSizex, self.imgSizey))

        grid = self.fig.add_gridspec(nrows=self.numRows, ncols=self.numColumns, left=0.05, right=0.48, wspace=0.05)
        axs = grid.subplots(sharex='col', sharey='row')

        for img_, tit_, img_grid_ in zip(self.video, self.list_titles, axs):

            if self.median_filter_flag:
                frame_v = median_filter(img_[int(frame), :, :], 3)
            else:
                frame_v = img_[int(frame), :, :]
            if self.correction_axis == 0:
                img_grid_.imshow(frame_v, cmap=self.color)
            elif self.correction_axis == 1:
                img_grid_.imshow(np.transpose(frame_v), cmap=self.color)
            if tit_ is not None:
                img_grid_.set_title(tit_)

        for ax in axs.flat:
            ax.label_outer()

        plt.show()

        fig = plt.figure(figsize=(self.imgSizex, self.imgSizey/10))
        grid = fig.add_gridspec(nrows=self.numRows, ncols=self.numColumns, left=0.05, right=0.48, wspace=0.05)
        axs = grid.subplots(sharex='col', sharey='row')

        for img_, img_grid_ in zip(self.video, axs):
            frame_v = img_[int(frame), :, :]
            gainMap1D_median = np.mean(frame_v, axis=self.correction_axis)

            x_grid = list(range(0, frame_v.shape[1]))
            y_grid = list(range(0, frame_v.shape[0]))

            x, y = np.meshgrid(x_grid, y_grid)
            if self.correction_axis == 0:
                img_grid_.plot(np.ravel(x), np.ravel(frame_v, order='A'), 'b.', linewidth=1, markersize=0.5)
            elif self.correction_axis == 1:
                img_grid_.plot(np.ravel(x), np.ravel(frame_v, order='F'), 'b.', linewidth=1, markersize=0.5)

            img_grid_.plot(gainMap1D_median, 'r-', linewidth=1, markersize=0.5)
            min_y = frame_v.mean() - 4*frame_v.std()
            max_y = frame_v.mean() + 4*frame_v.std()
            img_grid_.set_ylim(min_y, max_y)
            img_grid_.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))

        # Hide x labels and tick labels for all but bottom plot.
        for ax in axs.flat:
            ax.label_outer()

        plt.show()





class ImagingSetupParameters:

    ti0 = None
    ni0 = None
    ni = None
    tg0 = None
    tg = None
    ng0 = None
    ng = None
    ns = None
    wavelength = None
    k0 = None
    M = None
    NA = None
    alpha = None
    pixel_size_physical = None



from PySide6 import QtWidgets, QtCore, QtGui


class TemporalAnnotationSetting(QtWidgets.QWidget):
    temporalAnnotation_signal = QtCore.Signal(object)

    def __init__(self, parent=None):
        super(TemporalAnnotationSetting, self).__init__(parent)
        self.flag_smooth_filter = False
        self.window_size = 1

        # ok
        self.ok = QtWidgets.QPushButton("ok")
        self.ok.setAutoDefault(False)
        self.ok.clicked.connect(self.do_update)
        self.ok.setFixedWidth(100)

        self.le1 = QtWidgets.QLineEdit()
        self.le1.setPlaceholderText('Temporal window (#frames)')
        self.le_1_label = QtWidgets.QLabel("Temporal window:")
        self.le1.setFixedWidth(100)

        self.marker_size = QtWidgets.QLineEdit()
        self.marker_size.setPlaceholderText('Marker size (default=5))')
        self.marker_size_label = QtWidgets.QLabel("Marker size:")
        self.marker_size.setFixedWidth(100)

        self.checkbox_smooth_filter = QtWidgets.QCheckBox("Smoothness filter", self)

        self.grid = QtWidgets.QGridLayout()
        self.grid.addWidget(self.le_1_label, 0, 0)
        self.grid.addWidget(self.le1, 0, 1)
        self.grid.addWidget(self.marker_size_label, 1, 0)
        self.grid.addWidget(self.marker_size, 1, 1)
        self.grid.addWidget(self.checkbox_smooth_filter, 2, 0)
        self.grid.addWidget(self.ok, 3, 1)
        self.setLayout(self.grid)

        self.setWindowTitle("Temporal Annotation panel")
        self.setLayout(self.grid)

    @QtCore.Slot()
    def do_update(self):

        try:
            self.window_size = int(self.le1.text())
        except:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Incorrect input_video for step size! \nThe default value will be selected")
            self.msg_box.exec_()
            self.window_size = 1

        try:
            self.marker_size = int(self.marker_size.text())
        except:
            self.msg_box = QtWidgets.QMessageBox()
            self.msg_box.setWindowTitle("Warning!")
            self.msg_box.setText("Incorrect input_video for step size! \nThe default value will be selected")
            self.msg_box.exec_()
            self.marker_size = 5

        if self.checkbox_smooth_filter.isChecked():
            self.flag_smooth_filter = True
        else:
            self.flag_smooth_filter = False

        self.temporalAnnotation_signal.emit([self.window_size, self.flag_smooth_filter, self.marker_size])
        self.close()